from __future__ import annotations

import math
import typing as t
from dataclasses import KW_ONLY, dataclass

import typing_extensions as te
from uniserde import JsonDoc

import rio

from .fundamental_component import FundamentalComponent

__all__ = ["Grid"]


@t.final
@dataclass
class GridChildPosition:
    row: int
    column: int
    width: int = 1
    height: int = 1


@t.final
class Grid(FundamentalComponent):
    """
    A container which arranges its children in a table-like grid.

    Grids arrange their children in a table-like grid. Each child is placed in
    one or more cells of the grid. You can add children to the grid either by
    passing them into the constructor or by using the `Grid.add` method.

    To get full control over where children are placed, use `grid.add`. It
    returns the grid itself, so you can chain multiple `add` calls together for
    concise code.

    If you don't need all of that control, a convenient way of populating grids
    is by passing all children directly into the constructor. `Grid` accepts
    both individual components, as well as lists of components. Each value is
    interpreted as a single row of the grid, and the grid adjusted so that all
    rows fill the entire space evenly.


    ## Attributes

    `row_spacing`: The amount of space between rows of the grid.

    `column_spacing`: The amount of space between columns of the grid.


    ## Examples

    This code creates a grid layout with two rows and two columns, and adds
    children to the grid by passing them in as a list:

    ```python
    rio.Grid(
        [rio.Text("Hello"), rio.Text("World!")],  # 1. Row
        [rio.Text("Foo"), rio.Text("Bar")],  # 2. Row
    )
    ```

    Alternatively, you can use the `add` method to add children to the grid.
    Here's how you can do it:

    ```python
    grid = rio.Grid(row_spacing=1, column_spacing=1)
    grid.add(rio.Text("Hello"), row=0, column=0)
    grid.add(rio.Text("World!"), row=0, column=1)
    grid.add(rio.Text("Foo"), row=1, column=0)
    grid.add(rio.Text("Bar"), row=1, column=1)
    ```

    In your Component class, you can use the `add` method within the build function to
    add children to the grid. Here's how you can do it:

    ```python
    class MyComponent(rio.Component):
        def build(self) -> rio.Component:
            grid = rio.Grid(row_spacing=1, column_spacing=1)
            grid.add(rio.Text("Hello"), row=0, column=0)
            grid.add(rio.Text("World!"), row=0, column=1)
            grid.add(rio.Text("Foo"), row=1, column=0)
            grid.add(rio.Text("Bar"), row=1, column=1)

            return grid
    ```
    """

    _: KW_ONLY
    row_spacing: float
    column_spacing: float

    # These must be annotated, otherwise rio won't understand that grids have
    # child components and won't copy over the new values when two Grids are
    # reconciled
    _children: list[rio.Component]
    _child_positions: list[GridChildPosition]

    def __init__(
        self,
        *rows: rio.Component | t.Iterable[rio.Component],
        row_spacing: float = 0.0,
        column_spacing: float = 0.0,
        key: str | int | None = None,
        margin: float | None = None,
        margin_x: float | None = None,
        margin_y: float | None = None,
        margin_left: float | None = None,
        margin_top: float | None = None,
        margin_right: float | None = None,
        margin_bottom: float | None = None,
        min_width: float = 0,
        min_height: float = 0,
        # MAX-SIZE-BRANCH max_width: float | None = None,
        # MAX-SIZE-BRANCH max_height: float | None = None,
        grow_x: bool = False,
        grow_y: bool = False,
        align_x: float | None = None,
        align_y: float | None = None,
        # SCROLLING-REWORK scroll_x: t.Literal["never", "auto", "always"] = "never",
        # SCROLLING-REWORK scroll_y: t.Literal["never", "auto", "always"] = "never",
    ):
        super().__init__(
            key=key,
            margin=margin,
            margin_x=margin_x,
            margin_y=margin_y,
            margin_left=margin_left,
            margin_top=margin_top,
            margin_right=margin_right,
            margin_bottom=margin_bottom,
            min_width=min_width,
            min_height=min_height,
            # MAX-SIZE-BRANCH max_width=max_width,
            # MAX-SIZE-BRANCH max_height=max_height,
            grow_x=grow_x,
            grow_y=grow_y,
            align_x=align_x,
            align_y=align_y,
            # SCROLLING-REWORK scroll_x=scroll_x,
            # SCROLLING-REWORK scroll_y=scroll_y,
        )

        self.row_spacing = row_spacing
        self.column_spacing = column_spacing

        # JS can only work with lists of Components, so we'll store the
        # components and their positions separately
        self._children, self._child_positions = self._add_initial_children(rows)

        self._properties_set_by_creator_.update(
            ["_children", "_child_positions"]
        )

    def _add_initial_children(
        self,
        children: t.Iterable[rio.Component | t.Iterable[rio.Component]],
    ) -> tuple[list[rio.Component], list[GridChildPosition]]:
        """
        Adds the children added in the constructor to the component. This is
        fairly complex and thus has its own function so not to pollute the
        constructor.
        """
        result_children: list[rio.Component] = []
        result_child_positions: list[GridChildPosition] = []

        # Pre process the rows and find the number of children in each
        rows: list[list[rio.Component]] = []
        row_widths: list[int] = []

        for row in children:
            if isinstance(row, rio.Component):
                row = [row]
            else:
                row = list(row)

                # Don't die on empty rows
                if not row:
                    row = t.cast(
                        list[rio.Component],
                        [rio.Spacer(grow_x=False, grow_y=False)],
                    )

            rows.append(row)
            row_widths.append(len(row))

        # Find the target number of columns
        target_columns = math.lcm(*row_widths)
        assert target_columns > 0, (target_columns, row_widths)

        # Add the children
        for yy, row_components in enumerate(rows):
            row_width = row_widths[yy]
            multiplier = target_columns // row_width

            for xx, component in enumerate(row_components):
                result_children.append(component)
                result_child_positions.append(
                    GridChildPosition(
                        yy,
                        xx * multiplier,
                        width=multiplier,
                        height=1,
                    )
                )

        # Done
        return result_children, result_child_positions

    def add(
        self,
        child: rio.Component,
        row: int,
        column: int,
        *,
        width: int = 1,
        height: int = 1,
    ) -> te.Self:
        """
        Add a child to the grid at a specified position.

        Appends a child component to the end and then returns the
        `Grid`, which makes method chaining possible.


        ## Parameters

        `child`: The child component to add to the grid.

        `row`: The row in which to place the child.

        `column`: The column in which to place the child.

        `width`: The number of columns the child should take up.

        `height`: The number of rows the child should take up.


        ## Example

        ```python
        grid = rio.Grid(row_spacing=1, column_spacing=1)
        grid.add(rio.Text("Hello"), row=0, column=0)
        ```
        """
        assert isinstance(child, rio.Component), child

        if width <= 0:
            raise ValueError("Children have to take up at least one column")

        if height <= 0:
            raise ValueError("Children have to take up at least one row")

        self._children.append(child)
        self._child_positions.append(
            GridChildPosition(row, column, width, height)
        )

        # Return self for chaining
        return self

    def _custom_serialize_(self) -> JsonDoc:
        return {
            "_children": [child._id for child in self._children],
            "_child_positions": [vars(pos) for pos in self._child_positions],
        }


Grid._unique_id_ = "Grid-builtin"



from __future__ import annotations

import typing as t

import rio

from . import component

__all__ = ["DefaultRootComponent"]


OUTER_MARGIN = 0.8


def with_debug_tooltip(
    session: rio.Session,
    comp: component.Component,
    text: str,
) -> rio.Component:
    """
    Wraps the given component in a tooltip which only shows up if the app is
    running in debug mode. This allows you to provide additional information to
    users without disrupting their production site.
    """

    if session._app_server.debug_mode:
        return rio.Tooltip(
            anchor=comp,
            tip=text,
            position="right",
        )

    return comp


@t.final
class NavButton(component.Component):
    page: rio.ComponentPage
    is_current: bool

    def build(self) -> rio.Component:
        main_row = rio.Row()

        # Add a marker if this is the current page
        if self.is_current:
            marker = rio.Rectangle(
                fill=self.session.theme.primary_color,
                min_width=0.5,
                align_x=0,
                corner_radius=(0, 1, 1, 0),
            )
        else:
            marker = None

        main_row.add(
            rio.Switcher(
                marker,
                transition_time=0.1,
            )
        )

        # Prepare the base UI
        main_row.add(
            rio.Rectangle(
                content=rio.Link(
                    rio.Text(
                        self.page.name,
                        style=rio.TextStyle(
                            font_weight="bold" if self.is_current else "normal",
                        ),
                        selectable=False,
                        margin_x=OUTER_MARGIN,
                        margin_y=0.5,
                    ),
                    target_url=f"/{self.page.url_segment}",
                ),
                fill=rio.Color.TRANSPARENT,
                hover_fill=(
                    None
                    if self.is_current
                    else self.session.theme.neutral_palette.background_active
                ),
                ripple=True,
                transition_time=0.1,
                grow_x=True,
            )
        )

        return main_row


@t.final
class DefaultRootComponent(component.Component):
    """
    ## Metadata

    `public`: False
    """

    @rio.event.on_page_change
    async def _on_page_change(self) -> None:
        await self.force_refresh()

    def build(self) -> rio.Component:
        # Special case: If the app only has a single page, don't spawn any
        # navigation.
        if len(self.session.app.pages) == 1:
            return rio.PageView()

        # Regular case: Display a proper navigation
        #
        # Build up the main column
        main_column = rio.Column(
            spacing=0.5,
            margin_y=OUTER_MARGIN,
            min_width=20,
        )

        # App name
        main_column.add(
            with_debug_tooltip(
                session=self.session,
                comp=rio.Text(
                    self.session.app.name,
                    style="heading1",
                    overflow="wrap",
                    margin_left=OUTER_MARGIN,
                    # Leave some space for the curved corner
                    margin_right=OUTER_MARGIN
                    + self.session.theme.corner_radius_large,
                ),
                text="You can change the app name by passing `name=...` when creating your `rio.App` object.",
            )
        )

        # App description
        main_column.add(
            with_debug_tooltip(
                session=self.session,
                comp=rio.Text(
                    self.session.app.description,
                    overflow="wrap",
                    margin_x=OUTER_MARGIN,
                ),
                text="You can change the app description by passing `description=...` when creating your `rio.App` object.",
            )
        )

        # Which page is currently active?
        # Prepare the URL
        segments = self.session.active_page_url.parts

        # Special case: Not deep enough
        if len(segments) <= 1:
            # This won't match anything because it's not a valid URL segment"
            current_page_url = "///"
        else:
            current_page_url = segments[1]

        # Add navigation
        #
        # If this becomes any more involved, consider making it a separate
        # component
        pages = rio.Column(
            spacing=0.5,
            margin_y=1.5,
        )
        main_column.add(pages)

        for page in self.session.app.pages:
            if isinstance(page, rio.Redirect):
                continue

            pages.add(
                NavButton(
                    page,
                    is_current=page.url_segment == current_page_url,
                )
            )

        # Push the remaining UI to the bottom
        main_column.add(rio.Spacer())

        # Explain to the user how to get rid of this navigation
        if self.session._app_server.debug_mode:
            main_column.add(
                rio.Link(
                    "What's this?",
                    icon="material/library_books",
                    target_url="https://rio.dev/TODO/LINK/TO/HOWTO/REMOVE/NAVIGATION",
                    open_in_new_tab=True,
                    margin_x=OUTER_MARGIN,
                    margin_y=1,
                    align_x=0.5,
                )
            )

        # We are the coolest!
        main_column.add(
            rio.Link(
                rio.Row(
                    rio.Icon(
                        "rio/logo:color",
                        # fill="dim",
                        min_width=2,
                    ),
                    rio.Text(
                        "Built with Rio\nWebsites & apps in 100% Python",
                        justify="left",
                        style="dim",
                        grow_x=True,
                        # Leave some space for the curved corner
                        margin_right=self.session.theme.corner_radius_large,
                    ),
                    align_x=0.5,
                    spacing=0.5,
                ),
                target_url="https://rio.dev",
                margin_x=OUTER_MARGIN,
            )
        )

        # Build the UI
        return rio.Row(
            # Sidebar
            rio.Rectangle(
                content=main_column,
                fill=self.session.theme.neutral_color,
                corner_radius=(
                    0,
                    self.session.theme.corner_radius_large,
                    self.session.theme.corner_radius_large,
                    0,
                ),
                margin=0.5,
                margin_left=0,
            ),
            # The user's content
            rio.PageView(
                grow_x=True,
            ),
        )



import pytest

import rio.app_server


class FakeComponent(rio.Component):
    pass


class FakeSession:
    def __init__(self) -> None:
        self._base_url = rio.URL("http://foo.com")


PAGES = [
    rio.ComponentPage(
        name="Home",
        url_segment="",
        build=FakeComponent,
    ),
    rio.ComponentPage(
        name="Page 1",
        url_segment="page-1",
        build=FakeComponent,
    ),
    rio.ComponentPage(
        name="Page 2",
        url_segment="page-2",
        build=FakeComponent,
    ),
    rio.ComponentPage(
        name="Guard to Page 1",
        url_segment="guard-to-page-1",
        build=FakeComponent,
        guard=lambda _: "page-1",
    ),
    rio.ComponentPage(
        name="Guard to Cycle 1",
        url_segment="guard-to-cycle-1",
        build=FakeComponent,
        guard=lambda _: "guard-to-cycle-2",
    ),
    rio.ComponentPage(
        name="Guard to Cycle 2",
        url_segment="guard-to-cycle-2",
        build=FakeComponent,
        guard=lambda _: "guard-to-cycle-1",
    ),
    rio.Redirect(
        url_segment="redirect-to-page-1",
        target="page-1",
    ),
    rio.Redirect(
        url_segment="redirect-to-cycle-1",
        target="guard-to-cycle-2",
    ),
    rio.Redirect(
        url_segment="redirect-to-cycle-2",
        target="guard-to-cycle-1",
    ),
]


@pytest.mark.parametrize(
    "relative_url_before_redirects,relative_url_after_redirects_should",
    [
        # No redirects
        (
            "/page-1",
            "/page-1",
        ),
        (
            "/page-1?foo=bar",
            "/page-1?foo=bar",
        ),
        (
            "/page-1#foo",
            "/page-1#foo",
        ),
        # Redirects by guard
        (
            "/guard-to-page-1",
            "/page-1",
        ),
        (
            "/guard-to-page-1?foo=bar",
            "/page-1?foo=bar",
        ),
        (
            "/guard-to-page-1#foo",
            "/page-1#foo",
        ),
        # Redirects by `rio.Redirect`
        (
            "/redirect-to-page-1",
            "/page-1",
        ),
        (
            "/redirect-to-page-1?foo=bar",
            "/page-1?foo=bar",
        ),
        (
            "/redirect-to-page-1#foo",
            "/page-1#foo",
        ),
        # No such page
        (
            "/non-existent-page",
            "/non-existent-page",
        ),
        (
            "/non-existent-page?foo=bar",
            "/non-existent-page?foo=bar",
        ),
        (
            "/non-existent-page#foo",
            "/non-existent-page#foo",
        ),
        # TODO: Add tests for URLs with Unicode characters & escaped/unescaped
        # URLs
    ],
)
def test_redirects(
    relative_url_before_redirects: str,
    relative_url_after_redirects_should: str,
) -> None:
    # Create a fake session. It contains everything used by the routing system.
    # fake_session = t.cast(rio.Session, FakeSession())
    #
    # Apps only initialize their pages later, so the parameter here has no
    # effect. They are overwritten manually instead
    app = rio.App(
        pages=(),
    )
    app.pages = tuple(PAGES)

    fake_app_server = object.__new__(rio.app_server.FastapiServer)
    fake_app_server.app = app

    fake_session = object.__new__(rio.Session)
    fake_session._base_url = rio.URL("http://foo.com")
    fake_session._active_page_url = fake_session._base_url
    fake_session._app_server = fake_app_server

    # Determine the final URL
    active_pages, absolute_url_after_redirects_is = (
        rio.routing.check_page_guards(
            fake_session,
            fake_session._base_url.join(rio.URL(relative_url_before_redirects)),
        )
    )

    # Verify the final URL
    absolute_url_after_redirects_should = fake_session._base_url.join(
        rio.URL(relative_url_after_redirects_should)
    )

    assert (
        absolute_url_after_redirects_is == absolute_url_after_redirects_should
    )


test_redirects(
    "/guard-to-page-1",
    "/page-1",
)



import datetime
from dataclasses import dataclass

import rio


@dataclass
class TodoItem:
    title: str
    creation_time: datetime.datetime
    completed: bool = False


class TodoAppSettings(rio.UserSettings):
    todo_items: list[TodoItem] = []



import math
import typing as t

import pytest

import rio.data_models
import rio.debug.layouter
import rio.testing
from tests.utils.layouting import cleanup, verify_layout

# pytestmark = pytest.mark.async_timeout(30)


@pytest.fixture(scope="module", autouse=True)
async def cleanup_layouter():
    yield
    await cleanup()


@pytest.mark.parametrize(
    "text",
    [
        "",
        "short-text",
        "-".join(["long-text"] * 100),
    ],
)
async def test_single_component(text: str) -> None:
    """
    Just one component - this should fill the whole screen.
    """
    await verify_layout(lambda: rio.Text(text))


@pytest.mark.parametrize(
    "container_type",
    [rio.Row, rio.Column],
)
async def test_linear_container_with_no_extra_width(
    container_type: t.Type,
) -> None:
    await verify_layout(
        lambda: container_type(
            rio.Text("hi", min_width=100),
            rio.Button("clicky", min_width=400),
        )
    )


@pytest.mark.parametrize(
    "horizontal",
    [True, False],
)
@pytest.mark.parametrize(
    "first_child_grows",
    [True, False],
)
@pytest.mark.parametrize(
    "second_child_grows",
    [True, False],
)
@pytest.mark.parametrize(
    "proportions",
    [
        None,
        "homogeneous",
        [1, 2],
        [2, 1],
    ],
)
async def test_linear_container_with_extra_width(
    horizontal: bool,
    first_child_grows: bool,
    second_child_grows: bool,
    proportions: None | t.Literal["homogeneous"] | list[int],
) -> None:
    """
    A battery of scenarios to test the most common containers - Rows & Columns.
    """
    if horizontal:
        container_type = rio.Row

        if first_child_grows:
            first_child_width = 0
            first_child_grow_x = True
        else:
            first_child_width = 10
            first_child_grow_x = False

        if second_child_grows:
            second_child_width = 0
            second_child_grow_x = True
        else:
            second_child_width = 20
            second_child_grow_x = False

        first_child_height = 0
        second_child_height = 0

        first_child_grow_y = False
        second_child_grow_y = False

        parent_width = 50
        parent_height = 0
    else:
        container_type = rio.Column

        if first_child_grows:
            first_child_height = 0
            first_child_grow_y = True
        else:
            first_child_height = 10
            first_child_grow_y = False

        if second_child_grows:
            second_child_height = 0
            second_child_grow_y = True
        else:
            second_child_height = 20
            second_child_grow_y = False

        first_child_width = 0
        second_child_width = 0

        first_child_grow_x = False
        second_child_grow_x = False

        parent_width = 0
        parent_height = 50

    await verify_layout(
        lambda: container_type(
            rio.Text(
                "short-text",
                min_width=first_child_width,
                min_height=first_child_height,
                grow_x=first_child_grow_x,
                grow_y=first_child_grow_y,
            ),
            rio.Text(
                "very-much-longer-text",
                min_width=second_child_width,
                min_height=second_child_height,
                grow_x=second_child_grow_x,
                grow_y=second_child_grow_y,
            ),
            # It would be nice to vary the spacing as well, but that would once
            # again double the number of tests this case already has. Simply
            # always specify a spacing, since that is the harder case anyway.
            spacing=2,
            proportions=proportions,
            min_width=parent_width,
            min_height=parent_height,
            align_x=0.5,
            align_y=0.5,
        )
    )


async def test_stack() -> None:
    """
    All children in stacks should be the same size.
    """
    layouter = await verify_layout(
        lambda: rio.Stack(
            rio.Text("Small", key="small_text", min_width=10, min_height=20),
            rio.Text("Large", key="large_text", min_width=30, min_height=40),
            align_x=0,
            align_y=0,
        )
    )

    small_layout = layouter.get_layout_by_key("small_text")

    assert small_layout.left_in_viewport_inner == 0
    assert small_layout.top_in_viewport_inner == 0

    assert small_layout.allocated_inner_width == 30
    assert small_layout.allocated_inner_height == 40

    large_layout = layouter.get_layout_by_key("large_text")

    assert large_layout.left_in_viewport_inner == 0
    assert large_layout.top_in_viewport_inner == 0

    assert large_layout.allocated_inner_width == 30
    assert large_layout.allocated_inner_height == 40


@pytest.mark.parametrize(
    "parent_width,parent_height",
    [
        (10, 50),
        (50, 10),
    ],
)
async def test_aspect_ratio_container_small_child(
    parent_width: float,
    parent_height: float,
) -> None:
    """
    Create a rectangle with a fixed aspect ratio, and make sure it matches
    expectations.

    The child is smaller than the parent and will thus be pulled larger.
    """
    parent_aspect_ratio = parent_width / parent_height
    child_aspect_ratio = 2

    layout = await verify_layout(
        lambda: rio.Container(
            rio.AspectRatioContainer(
                rio.Rectangle(
                    fill=rio.Color.RED,
                    key="child",
                ),
                aspect_ratio=child_aspect_ratio,
            ),
            min_width=parent_width,
            min_height=parent_height,
            align_x=0,
            align_y=0,
        )
    )

    # How large should the child be?
    if parent_aspect_ratio > child_aspect_ratio:
        child_width_should = parent_height * child_aspect_ratio
        child_height_should = parent_height
    else:
        child_width_should = parent_width
        child_height_should = parent_width / child_aspect_ratio

    # Is it though?
    child_layout = layout.get_layout_by_key("child")

    assert math.isclose(child_layout.allocated_inner_width, child_width_should)
    assert math.isclose(
        child_layout.allocated_inner_height, child_height_should
    )


@pytest.mark.parametrize(
    "child_specified_width,child_specified_height,child_width_should,child_height_should",
    [
        (10, 50, 20, 100),
        (50, 10, 50, 25),
    ],
)
async def test_aspect_ratio_container_large_child(
    child_specified_width: float,
    child_specified_height: float,
    child_width_should: float,
    child_height_should: float,
) -> None:
    """
    Create a rectangle with a fixed aspect ratio, and make sure it matches
    expectations.

    The child is larger than the parent and will thus push the parent larger.
    """
    child_aspect_ratio = 2

    layout = await verify_layout(
        lambda: rio.Container(
            rio.AspectRatioContainer(
                rio.Rectangle(
                    fill=rio.Color.RED,
                    min_width=child_specified_width,
                    min_height=child_specified_height,
                    key="child",
                ),
                aspect_ratio=child_aspect_ratio,
            ),
            min_width=20,
            min_height=20,
            align_x=0,
            align_y=0,
        )
    )

    # Is it though?
    child_layout = layout.get_layout_by_key("child")

    assert math.isclose(child_layout.allocated_inner_width, child_width_should)
    assert math.isclose(
        child_layout.allocated_inner_height, child_height_should
    )


@pytest.mark.parametrize(
    "scroll_x,scroll_y",
    [
        ("never", "auto"),
        ("auto", "never"),
        ("auto", "auto"),
    ],
)
async def test_scrolling(
    scroll_x: t.Literal["never", "always", "auto"],
    scroll_y: t.Literal["never", "always", "auto"],
) -> None:
    await verify_layout(
        lambda: rio.ScrollContainer(
            rio.Text("hi", min_width=30, min_height=30),
            scroll_x=scroll_x,
            scroll_y=scroll_y,
            min_width=20,
            min_height=20,
            align_x=0.5,
            align_y=0.5,
        )
    )


async def test_ellipsized_text() -> None:
    layouter = await verify_layout(
        lambda: rio.Text(
            "My natural size should become 0",
            overflow="ellipsize",
            align_x=0,
            key="text",
        )
    )

    layout = layouter.get_layout_by_key("text")

    assert layout.natural_width == 0


@pytest.mark.parametrize(
    "justify",
    [
        "left",
        "right",
        "center",
        "justified",
        "grow",
    ],
)
async def test_flow_container_layout(justify: str) -> None:
    await verify_layout(
        lambda: rio.FlowContainer(
            rio.Text("foo", min_width=5),
            rio.Text("bar", min_width=10),
            rio.Text("qux", min_width=4),
            column_spacing=3,
            row_spacing=2,
            justify=justify,  # type: ignore
            min_width=20,
            align_x=0,
        )
    )



from __future__ import annotations

import typing as t
from dataclasses import KW_ONLY
from pathlib import Path

from uniserde import JsonDoc

from .. import color, fills, icon_registry
from .fundamental_component import FundamentalComponent

__all__ = [
    "Icon",
]


_IconFill = t.Union[
    "fills.SolidFill",
    "fills.LinearGradientFill",
    "fills.ImageFill",
    "color.ColorSet",
    t.Literal["dim"],
]


@t.final
class Icon(FundamentalComponent):
    """
    Displays one of many pre-bundled icons.

    Icons are a great way to add polish to your app. A good icon can help your
    users understand your app and immediately recognize what a component does.

    Rio includes hundreds of free icons, allowing you to easily add them to your
    app without having to find or create your own. The `Icon` component displays
    one of these icons.

    Note that unlike most components in Rio, the `Icon` component does not have
    a `natural` size, since icons can be easily be scaled to fit any space.
    Because of this, `Icon` defaults to a width and height of 1.3, which is a
    great size when mixing icons with text.

    Icon names are in the format `icon_set/icon_name:variant`. Rio already ships
    with the `material` icon set, which contains icons in the style of Google's
    Material Design. You can browse all available icons in Rio's dev tools. (The
    dev tools sidebar is visible on the right-hand-side when running your
    project using `rio run`.)

    The set name and variant can be omitted. If no set name is specified, it
    defaults to `material`. If no variant is specified, the default version of
    the icon, i.e. no variant, is used.


    ## Attributes

    `icon`: The name of the icon to display, in the format
        `icon_set/icon_name:variant`. You can browse all available icons in
        Rio's dev tools sidebar.

    `fill`: The color scheme of the icon. The text color is used if no fill is
        specified.


    ## Examples

    This minimal example will display the icon named "castle" from the
    "material" icon set:

    ```python
    rio.Icon("material/castle")
    ```

    You can also specify the color, width and height of the icon:

    ```python
    rio.Icon(
        "material/castle",
        fill=rio.Color.from_hex("ff0000"),
        min_height=2.5,
        min_width=2.5,
    )
    ```
    """

    icon: str
    _: KW_ONLY
    fill: _IconFill

    @staticmethod
    def register_icon_set(
        set_name: str,
        set_archive_path: Path,
    ) -> None:
        """
        Adds an icon set to the global registry. This allows the icons to be
        accessed as `"icon_set/icon_name"` or `"icon_set/icon_name:variant"`.

        There must not already be a set with the given name.

        The icon set is a `.tar.xz` compressed archive and must contain exactly
        one directory, which must be named identically to the icon set. Files
        located in the root of that directory can be accessed as
        `"icon_set/icon_name"`. Files located in a subdirectory can be accessed
        as `"icon_set/icon_name:variant"`.

        For SVG files to work as icons...

        - They must have a `viewBox` attribute, but no height or width
        - They must contain exactly one XML root node: `<svg>...</svg>`.
          Comments also count!
        - Rio colors paths by assigning a `fill` to the SVG root. This only
          works as long as the SVG paths don't have a `<style>` assigned already.


        ## Parameters

        `set_name`: The name of the new icon set. This will be used to access
            the icons.

        `set_archive_path`: The path to the `.tar.xz` archive containing the
            icon set.
        """

        icon_registry.register_icon_set(set_name, set_archive_path)

    @staticmethod
    def register_single_icon(
        icon_source: Path,
        set_name: str,
        icon_name: str,
        variant_name: str | None = None,
    ) -> None:
        """
        Adds a single icon to the global registry. This allows the icon to be
        accessed as `"icon_set/icon_name"` or `"icon_set/icon_name:variant"`.

        `icon_source` needs to be the path to a single SVG file. For SVG files
        to work as icons, they need to meet some conventions:

        - They must have a `viewBox` attribute, but no height or width
        - They must contain exactly one XML root node: `<svg>...</svg>`.
          Comments also count!
        - Rio colors paths by assigning a `fill` to the SVG root. This only
        works as long as the SVG paths don't have a `<style>` assigned
        already.


        ## Parameters

        `icon_source`: The path to the SVG file containing the icon.

        `set_name`: The name of the new icon set. This will be used to access
            the icons.

        `icon_name`: The name of the icon. This will be used to access the
            icon.

        `variant_name`: The name of the variant. This will be used to access
            the icon. If not specified, the default variant will be used.
        """

        icon_registry.register_single_icon(
            icon_source, set_name, icon_name, variant_name
        )

    def __init__(
        self,
        icon: str,
        *,
        fill: _IconFill = "keep",
        key: str | int | None = None,
        margin: float | None = None,
        margin_x: float | None = None,
        margin_y: float | None = None,
        margin_left: float | None = None,
        margin_top: float | None = None,
        margin_right: float | None = None,
        margin_bottom: float | None = None,
        min_width: float = 1.3,
        min_height: float = 1.3,
        # MAX-SIZE-BRANCH max_width: float | None = None,
        # MAX-SIZE-BRANCH max_height: float | None = None,
        grow_x: bool = False,
        grow_y: bool = False,
        align_x: float | None = None,
        align_y: float | None = None,
        # SCROLLING-REWORK scroll_x: t.Literal["never", "auto", "always"] = "never",
        # SCROLLING-REWORK scroll_y: t.Literal["never", "auto", "always"] = "never",
    ):
        super().__init__(
            key=key,
            margin=margin,
            margin_x=margin_x,
            margin_y=margin_y,
            margin_left=margin_left,
            margin_top=margin_top,
            margin_right=margin_right,
            margin_bottom=margin_bottom,
            min_width=min_width,
            min_height=min_height,
            # MAX-SIZE-BRANCH max_width=max_width,
            # MAX-SIZE-BRANCH max_height=max_height,
            grow_x=grow_x,
            grow_y=grow_y,
            align_x=align_x,
            align_y=align_y,
            # SCROLLING-REWORK scroll_x=scroll_x,
            # SCROLLING-REWORK scroll_y=scroll_y,
        )

        self.icon = icon
        self.fill = fill

    def __post_init__(self) -> None:
        # Verify that the icon exists. This makes sure any crashes happen
        # immediately, rather than during the next refresh.
        icon_registry.get_icon_svg(self.icon)

    def _custom_serialize_(self) -> JsonDoc:
        # Serialize the fill. This isn't automatically handled because it's a
        # Union.
        fill = self.fill
        if not isinstance(fill, str):
            fill = self.session._serialize_fill(fill)

        # Serialize
        return {
            "fill": fill,
        }


Icon._unique_id_ = "Icon-builtin"



import gc
import random
import asyncio
import json
import rio
import rio.components.class_container
import rio.data_models
import rio.debug
import rio.debug.dev_tools
import rio.debug.dev_tools.dev_tools_connector
import rio.debug.dev_tools.icons_page
import rio.debug.dev_tools.layout_display
import rio.debug.layouter


def filter_function(component: rio.Component) -> bool:
    # Don't care about the connection lost popup
    # if type(component).__name__ == "DefaultConnectionLostComponent":
    #     return False

    # Everything else is fine
    return True


class LoginBox(rio.Component):
    TEXT_STYLE = rio.TextStyle(fill=rio.Color.from_hex("02dac5"), font_size=0.9)

    def build(self) -> rio.Component:
        return rio.Rectangle(
            content=rio.Column(
                rio.TextInput(
                    text="",
                    label="Benutzername",
                    accessibility_label="Benutzername",
                    min_height=0.5,
                ),
                rio.TextInput(
                    text="",
                    label="Passwort",
                    accessibility_label="Passwort",
                    is_secret=True,
                ),
                rio.Column(
                    rio.Row(
                        rio.Button(
                            rio.Text("LOGIN", style=self.TEXT_STYLE),
                            shape="rectangle",
                            style="minor",
                            color="secondary",
                            margin_bottom=0.4,
                        )
                    ),
                    rio.Row(
                        rio.Button(
                            rio.Text("REG", style=self.TEXT_STYLE),
                            shape="rectangle",
                            style="minor",
                            color="secondary",
                        ),
                        rio.Spacer(),
                        rio.Button(
                            rio.Text("LST PWD", style=self.TEXT_STYLE),
                            shape="rectangle",
                            style="minor",
                            color="secondary",
                        ),
                        proportions=(49, 2, 49),
                    ),
                ),
                spacing=0.4,
            ),
            fill=rio.Color.TRANSPARENT,
            align_x=0.5,
            align_y=0.5,
        )


class MyRoot(rio.Component):
    messages: list[str] = [
        "Initial",
    ]
    text_value: str = "<->"
    number_value: float = 0

    is_on: bool = False

    entries: list[int] = list(range(10))

    async def on_randomize(self) -> None:
        random.shuffle(self.entries)
        await self.force_refresh()

    async def _find_dead_components(self) -> None:
        try:
            import objgraph  # type: ignore
        except ImportError:
            print("Please install `objgraph`")
            return
        key = ("Foobar",)
        # Make sure only real problem components exist
        gc.collect()

        # Find all components present in the app
        alive_component_ids: set[int] = set()
        to_do: list[rio.Component] = [self.session._root_component]

        while to_do:
            component = to_do.pop()
            alive_component_ids.add(id(component))
            to_do.extend(component._iter_direct_children_())

        # Find all components which are alive according to Python, but aren't
        # part of the app
        zombie_component_ids: set[int] = (
            set(self.session._weak_components_by_id.keys())
            - alive_component_ids
        )

        # Select one component to debug
        component = None

        for cmp_id, component in self.session._weak_components_by_id.items():
            if cmp_id not in zombie_component_ids:
                continue

            if isinstance(component, rio.debug.dev_tools.icons_page.IconsPage):
                gc.collect()
                objgraph.show_backrefs(
                    [component],
                    filename="/home/jakob/Downloads/sample-graph.png",
                    max_depth=10,
                )

    def build_dialog(self) -> rio.Component:
        return rio.Card(
            rio.Button(
                "Hello, World!",
                on_press=lambda: print("Hello, World!"),
                margin=3,
            ),
            align_x=0.5,
            align_y=0.5,
        )

    async def on_press(self) -> None:
        self.is_on = not self.is_on

    async def on_button_press(self) -> None:
        await self.session.set_clipboard("Hello, World!")

    def build(self) -> rio.Component:
        return rio.Popup(
            anchor=rio.Button(
                "Toggle Popup",
                on_press=self.on_press,
            ),
            content=rio.Text("I'm here!"),
            is_open=self.is_on,
            # is_open=True,
            align_x=0.5,
            align_y=0.5,
            position="top",
        )

        table = rio.Table(
            # pd.DataFrame(
            {
                "a": [1, 2, 3],
                "b": [4, 5, 6],
                "c": [7, 8, 9],
                "d": [10, 11, 12],
            }
            # ),
        )

        table["header", -2:].style(
            font_weight="bold",
        )

        table[:2, 0].style(
            font_weight="bold",
        )

        return table

        return rio.Column(
            # blabla
            *([] if self.content is None else [self.content])
            # blabla
        )

        return rio.Column(
            # blabla
            rio.Switcher(self.content),
            # blabla
        )

        return rio.Button(
            "Open Dialog",
            on_press=self.on_press,
            align_x=0.5,
            align_y=0.5,
        )

        return rio.Row(
            rio.Popup(
                anchor=rio.Button(
                    "Toggle Popup",
                    on_press=lambda: setattr(self, "is_on", not self.is_on),
                ),
                content=rio.Button(
                    "Toggle Popup",
                    on_press=lambda: setattr(self, "is_on", not self.is_on),
                    margin=1,
                ),
                position="left",
                gap=10,
                # alignment=1,
                is_open=self.is_on,
                align_x=0.5,
                align_y=0.5,
            ),
            rio.Tooltip(
                anchor=rio.Text("Hover for Tooltip"),
                tip=rio.Text("This is a tooltip"),
                align_x=0.5,
                align_y=0.5,
            ),
            spacing=3,
        )

        return rio.Column(
            rio.FlowContainer(
                *[
                    rio.Text(
                        f"Entry {ii}",
                        key=f"entry_{ii}",
                        margin=0.5,
                    )
                    for ii in self.entries
                ],
                grow_y=True,
            ),
            rio.Button(
                "Randomize",
                on_press=self.on_randomize,
            ),
        )


ALL_IMAGES = [
    rio.URL(
        "https://fastly.picsum.photos/id/13/2500/1667.jpg?hmac=SoX9UoHhN8HyklRA4A3vcCWJMVtiBXUg0W4ljWTor7s"
    ),
    rio.URL(
        "https://fastly.picsum.photos/id/19/2500/1667.jpg?hmac=7epGozH4QjToGaBf_xb2HbFTXoV5o8n_cYzB7I4lt6g"
    ),
    rio.URL(
        "https://fastly.picsum.photos/id/28/4928/3264.jpg?hmac=GnYF-RnBUg44PFfU5pcw_Qs0ReOyStdnZ8MtQWJqTfA"
    ),
    rio.URL(
        "https://fastly.picsum.photos/id/29/4000/2670.jpg?hmac=rCbRAl24FzrSzwlR5tL-Aqzyu5tX_PA95VJtnUXegGU"
    ),
]


class MyRoot(rio.Component):
    def _on_file(self, file: rio.FileInfo) -> None:
        print(f"Got file {file.name} with size {file.size_in_bytes}")

    def build(self) -> rio.Component:
        return rio.Column(
            rio.TextInput(
                text="Foobar",
            ),
            rio.TextInput(
                text="Foobar",
                style="rounded",
            ),
            rio.TextInput(
                text="Foobar",
                style="pill",
            ),
            rio.Text(
                "Foobar",
                justify="center",
                style=rio.TextStyle(),
            ),
            rio.Text(
                "Foobar",
                justify="center",
                style=rio.TextStyle(
                    underlined=True,
                ),
            ),
            rio.Text(
                "Foobar",
                justify="center",
                style=rio.TextStyle(
                    strikethrough=True,
                ),
            ),
            rio.Text(
                "Foobar",
                justify="center",
                style=rio.TextStyle(
                    underlined=True,
                    strikethrough=True,
                ),
            ),
            spacing=0.5,
            min_width=10,
            align_x=0.5,
            align_y=0.5,
        )

        return rio.ColorPicker(
            color=rio.Color.RED,
        )

        file_types = [
            # "pdf",
            # "doc",
            # "jpeg",
            # "mp3",
            "ogg",
            "mkv",
        ]

        # return rio.Column(
        #     rio.FilePickerArea(
        #         file_types=file_types,
        #         on_file_upload=self._on_file,
        #     ),
        #     rio.FilePickerArea(
        #         file_types=file_types,
        #         on_file_upload=self._on_file,
        #         min_width=80,
        #         min_height=30,
        #     ),
        #     spacing=8,
        #     align_x=0.5,
        #     align_y=0.5,
        # )

        return rio.Button(
            "Open Dialog",
            style="colored-text",
            on_press=lambda: self.session.show_yes_no_dialog(
                "Heyho?",
                owning_component=self,
            ),
            align_x=0.5,
            align_y=0.35,
        )


class ThemeSample(rio.Component):
    theme: rio.Theme

    async def _apply_theme_worker(self) -> None:
        # Wait for a bit to ensure the client has time to update the HTML
        await asyncio.sleep(0.2)

        # Get all variables to apply
        light_or_dark = "light" if self.theme.is_light_theme else "dark"
        theme_vars = self.session._calculate_theme_css_values(self.theme)

        # Apply the theme
        await self.session._evaluate_javascript(
            f"""
let themeVars = {json.dumps(theme_vars)};

let elem = globalThis.componentsById[{self._id}].element;

for (let key in themeVars) {{
    elem.style.setProperty(
        key,
        themeVars[key],
    );
}}

// Set the theme variant
document.documentElement.setAttribute(
    "data-theme",
    {json.dumps(light_or_dark)}
);

            """
        )

    def build(self) -> rio.Component:
        # Apply the theme
        asyncio.create_task(
            self._apply_theme_worker(),
        )

        # Build the content
        content = rio.Row(
            # Main Content
            rio.Column(
                rio.Text(
                    "Lorem Ipsum",
                    style="heading1",
                ),
                rio.Markdown(
                    """
Lorem ipsum dolor sit amet, consectetur adipiscing elit. Nullam nec
ultricies elit. Nullam nec ultricies elit. Nullam nec ultricies elit.
                    """,
                ),
                rio.Image(
                    rio.URL("https://picsum.photos/300/200"),
                    min_width=26,
                    min_height=14,
                    corner_radius=self.theme.corner_radius_medium,
                ),
                rio.Markdown(
                    """
Lorem markdownum auctor fiant aversa, scopulos bos ima ruptosque turbam, fertur
flexile amictu temptaretque *superum temptarunt* volatu ad veros. Fatebor
Aesacos.

```js
rom_root(record);
direct = subdirectory(symbolic_restore_t, url);
```

Coniuge et tamen et omnibus ille concordant mitissima accedere fratrem
auditurum, spes petunt ore insigne mille. Te explorant ardor, dies inde eloquio
rostrum vestemque et frustra purum. Cynthia quoniam dextra aequoris, emissi [in
                    """,
                ),
                grow_x=True,
            ),
            # (Fake) Drawer
            rio.ThemeContextSwitcher(
                content=rio.Rectangle(
                    content=rio.Grid(
                        rio.Text(
                            "Drawer Heading",
                            style="heading2",
                        ),
                        rio.TextInput(
                            text="Foobar",
                            style="rounded",
                        ),
                        [
                            rio.Text("Switch"),
                            rio.Switch(),
                        ],
                        [
                            rio.Button("Major", style="major"),
                            rio.Button("Minor", style="minor"),
                        ],
                        [
                            rio.Button("Colored Text", style="colored-text"),
                            rio.Button("Plain Text", style="plain-text"),
                        ],
                        rio.Spacer(),
                        rio.Separator(),
                        rio.Row(
                            rio.Icon("material/castle"),
                            rio.Text(
                                "Castles are Cool",
                                grow_x=True,
                            ),
                            spacing=0.5,
                        ),
                        row_spacing=1,
                        column_spacing=1,
                        margin=2,
                    ),
                    fill=self.theme.neutral_color,
                    corner_radius=(
                        self.theme.corner_radius_large,
                        0,
                        0,
                        self.theme.corner_radius_large,
                    ),
                    shadow_color=self.theme.shadow_color,
                    shadow_radius=1.3,
                ),
                color="neutral",
            ),
            spacing=1,
        )

        # Wrap up the content
        #
        # A rectangle acts as a fake background
        return rio.Rectangle(
            # A Theme context switcher ensures that the theme variables are
            # locally exposed.
            content=rio.ThemeContextSwitcher(
                content=content,
                color="background",
            ),
            fill=self.theme.background_color,
        )


class ThemeViewer(rio.Component):
    def build(self) -> rio.Component:
        # Decide on the colors to use
        theme_colors: list[rio.Color] = [
            self.session.theme.primary_color,
            rio.Color.RED,
            rio.Color.GREEN,
            rio.Color.YELLOW,
        ]

        # Create the themes
        light_viewers: list[ThemeSample] = []
        dark_viewers: list[ThemeSample] = []

        for col in theme_colors:
            light_theme, dark_theme = rio.Theme.pair_from_colors(
                primary_color=col,
            )

            light_viewers.append(ThemeSample(theme=light_theme))
            dark_viewers.append(ThemeSample(theme=dark_theme))

        # Populate
        return rio.Grid(
            light_viewers,
            dark_viewers,
            row_spacing=0.4,
            column_spacing=0.4,
        )


app = rio.App(
    # Uncomment this line to enable the custom navigation
    build=ThemeViewer,
    theme=rio.Theme.pair_from_colors(),
    default_attachments=[],
)



from __future__ import annotations

import rio


# <component>
@rio.page(
    name="About Us",
    url_segment="about-page",
)
class AboutPage(rio.Component):
    """
    A sample page, which displays a humorous description of the company.
    """

    def build(self) -> rio.Component:
        return rio.Markdown(
            """
# About Us

Welcome to Buzzwordz Inc.! Unleashing Synergistic Paradigms for Unprecedented
Excellence since the day after yesterday.

## About Our Company

At buzzwordz, we are all talk and no action. Our mission is to be the vanguards
of industry-leading solutions, leveraging bleeding-edge technologies to catapult
your business into the stratosphere of success. Our unparalleled team of ninjas,
gurus, and rockstars is dedicated to disrupting the status quo and actualizing
your wildest business dreams. We live, breathe, and eat operational excellence
and groundbreaking innovation.

## Synergistic Consulting

Unlock your business's quantum potential with our bespoke, game-changing
strategies. Our consulting services synergize cross-functional paradigms to
create a holistic ecosystem of perpetual growth and exponential ROI. Did I
mention paradigm-shifts? We've got those too.

## Agile Hyper-Development

We turn moonshot ideas into reality with our agile, ninja-level development
techniques. Our team of coding wizards crafts robust, scalable, and future-proof
solutions that redefine industry standards. 24/7 Proactive Hyper-Support

Experience next-gen support that anticipates your needs before you do. Our
omnipresent customer happiness engineers ensure seamless integration,
frictionless operation, and infinite satisfaction, day and night.

Embark on a journey of transformational growth and stratospheric success. Don't
delay, give us your money today.

Phone: (123) 456-7890

Email: info@yourwebsite.com

Address: 123 Main Street, City, Country
            """,
            min_width=60,
            margin_bottom=4,
            align_x=0.5,
            align_y=0,
        )


# </component>



import pytest

import rio


@pytest.mark.parametrize(
    "input_type,output_type",
    [
        # Simple cases
        ("pdf", "pdf"),
        (".Pdf", "pdf"),
        (".PDF", "pdf"),
        ("*pdf", "pdf"),
        ("*.Pdf", "pdf"),
        ("*.PDF", "pdf"),
        ("application/pdf", "pdf"),
        # Make the results are standardized
        (".jpg", "jpg"),
        (".jpeg", "jpg"),
        # Invalid MIME types
        ("not/a/real/type", "type"),
        ("////Type", "type"),
    ],
)
def test_standardize_file_types(
    input_type: str,
    output_type: str,
) -> None:
    cleaned_type = rio.utils.normalize_file_type(input_type)
    assert cleaned_type == output_type


@pytest.mark.parametrize(
    "unsorted_sequence,keys,sorted_sequence_should",
    [
        (
            [0, 1, 2, 3, 4],
            [0, 1, 2, 3, 4],
            [0, 1, 2, 3, 4],
        ),
        (
            [4, 3, 2, 1, 0],
            [4, 3, 2, 1, 0],
            [0, 1, 2, 3, 4],
        ),
        (
            [4, 3, 2, 1, 0],
            [4, None, None, None, 0],
            [0, 3, 2, 1, 4],
        ),
        (
            [0, 1, 2, 3],
            [0, 1, 1, 3],
            [0, 1, 2, 3],
        ),
        (
            [3, 2, 1, 0],
            [3, 1, 1, 0],
            [0, 2, 1, 3],
        ),
        (
            [0, 1, 2, 3],
            [0, None, None, None],
            [0, 1, 2, 3],
        ),
        (
            [0, 1, 2, 3],
            [9, None, None, None],
            [1, 2, 3, 0],
        ),
    ],
)
def test_soft_sort(
    unsorted_sequence: list[int],
    keys: list[int],
    sorted_sequence_should: list[int],
) -> None:
    # Sort the sequence
    sorted_sequence_is = unsorted_sequence.copy()
    rio.utils.soft_sort(
        sorted_sequence_is,
        key=lambda x: keys[unsorted_sequence.index(x)],
    )

    # Verify the result
    assert sorted_sequence_is == sorted_sequence_should



from __future__ import annotations

import random
import typing as t
from dataclasses import KW_ONLY
from datetime import date

import rio

from .. import components as comps
from .component import Component

__all__ = [
    "DateInput",
]


def make_fake_input_box(
    *,
    theme: rio.Theme,
    label: str,
    value: str,
) -> rio.Component:
    palette = theme.neutral_palette

    label_style = rio.TextStyle(
        fill=theme.secondary_color,
        font_size=0.8,
    )

    return rio.Column(
        # The background rectangle
        rio.Rectangle(
            content=rio.Column(
                rio.Text(
                    label,
                    selectable=False,
                    style=label_style,
                ),
                rio.Row(
                    rio.Text(
                        value,
                        justify="left",
                        selectable=False,
                        margin_bottom=0.4,
                        align_y=1,
                        grow_x=True,
                    ),
                    rio.Icon(
                        "material/calendar_today:fill",
                        fill="dim",
                        min_width=1.5,
                        min_height=1.5,
                        margin_bottom=0.3,
                        align_y=1,
                    ),
                    spacing=0.8,
                    grow_y=True,
                ),
                # spacing=0.2,
                margin_x=1,
                margin_top=0.5,
            ),
            fill=palette.background,
            hover_fill=palette.background_active,
            corner_radius=(
                theme.corner_radius_small,
                theme.corner_radius_small,
                0,
                0,
            ),
            cursor=rio.CursorStyle.POINTER,
            grow_y=True,
            transition_time=0.1,
        ),
        # The line at the bottom
        rio.Rectangle(
            fill=palette.foreground.replace(opacity=0.25),
            min_height=0.12,
        ),
        min_width=9,
    )


@t.final
class DateInput(Component):
    """
    Allows the user to pick a date from a calendar.

    DateInputs are similar in appearance to `TextInput` and `NumberInput`, but
    allow the user to pick a date from a calendar, rather than text or number.
    When pressed, a calendar will pop-up, allowing the user to select a date.

    This makes for a compact component, which still allows the user to visually
    select a date.

    For a larger component which permanently displays a calendar, consider using
    `rio.Calendar`.


    ## Attributes

    `value`: The currently selected date.

    `label`: A short text to display next to the input field.

    `on_change`: Triggered whenever the user selects a new date.


    ## Examples

    Here's a simple example that allows the user to select a data and displays
    it back to them:

    ```python
    from datetime import date

    class MyComponent(rio.Component):
        value: date = date.today()

        def build(self) -> rio.Component:
            return rio.Column(
                rio.DateInput(
                    # In order to retrieve a value from the component, we'll
                    # use an attribute binding. This way our own value will
                    # be updated whenever the user changes the value.
                    value=self.bind().value,
                    label="Pick a Date",
                ),
                rio.Text(f"You've selected: {self.value}"),
            )
    ```

    Alternatively you can also attach an event handler to react to changes. This
    is a little more verbose, but allows you to run arbitrary code when the user
    picks a new date:

    ```python
    from datetime import date

    class MyComponent(rio.Component):
        value: date = date.today()

        def on_value_change(self, event: rio.DateChangeEvent):
            # This function will be called whenever the input's value
            # changes. We'll display the new value in addition to updating
            # our own attribute.
            self.value = event.value
            print(f"You've selected: {self.value}")

        def build(self) -> rio.Component:
            return rio.DateInput(
                value=self.value,
                label="Pick a Date",
                on_change=self.on_value_change,
            )
    ```


    ## Metadata

    `experimental`: True
    """

    value: date

    _: KW_ONLY

    label: str = ""
    on_change: rio.EventHandler[rio.DateChangeEvent] = None

    _is_open: bool = False

    async def _on_value_change(self, event: rio.DateChangeEvent) -> None:
        # Close the date picker
        self._is_open = False

        # Chain the event handler
        await self.call_event_handler(self.on_change, event)

    def _on_toggle_open(self, _: rio.PressEvent) -> None:
        self._is_open = not self._is_open

    def _on_close(self) -> None:
        self._is_open = False

    def build(self) -> rio.Component:
        return rio.Popup(
            # Place a fake textbox. It's only used for styling and displaying
            # the label, if any
            anchor=rio.MouseEventListener(
                content=make_fake_input_box(
                    theme=self.session.theme,
                    label=self.label,
                    value=self.value.strftime(self.session._date_format_string),
                ),
                on_press=self._on_toggle_open,
            ),
            # Display a calendar so the user can pick a date
            content=rio.Column(
                comps.Calendar(
                    value=self.bind().value,
                    on_change=self._on_value_change,
                    # Use a fresh calendar each time, so it displays the
                    # currently selected value each time the popup opens
                    key=str(random.random()),
                ),
                rio.Button(
                    "Cancel",
                    style="colored-text",
                    on_press=self._on_close,
                ),
                margin=1,
            ),
            color="neutral",
            position="center",
            alignment=0.5,
            is_open=self._is_open,
        )



from __future__ import annotations

import logging
import typing as t
import warnings
from dataclasses import KW_ONLY, dataclass, field
from pathlib import Path

import introspection
from introspection import convert_case

import rio
import rio.docs

from . import deprecations, utils
from .errors import NavigationFailed

__all__ = [
    "Redirect",
    "ComponentPage",
    "Page",
    "page",
    "GuardEvent",
]


DEFAULT_ICON = "rio/logo:color"


@t.final
@dataclass(frozen=True)
class Redirect:
    """
    Redirects the user to a different page.

    Redirects can be added to the app in place of "real" pages. These are
    useful for cases where you want to add valid links to the app, but don't
    want to display a page at that URL. They redirect users to another page
    the instant they would be opened.

    Redirects are passed directly to the app during construction, like so:

    ```python
    import rio

    app = rio.App(
        build=lambda: rio.Column(
            rio.Text("Welcome to my app!"),
            rio.PageView(grow_y=True),
        ),
        pages=[
            rio.ComponentPage(
                name="Home",
                url_segment="",
                build=lambda: rio.Text("This is the home page"),
            ),
            rio.ComponentPage(
                name="Subpage",
                url_segment="subpage",
                build=lambda: rio.Text("This is a subpage"),
            ),
            rio.Redirect(
                url_segment="old-page",
                target="/subpage",
            ),
        ],
    )

    app.run_in_browser()
    ```
    """

    url_segment: str
    target: str | rio.URL


@t.final
@dataclass(frozen=True)
class ComponentPage:
    """
    A routable page in a Rio app.

    Rio apps can consist of many pages. You might have a welcome page, a
    settings page, a login, and so on. `ComponentPage` components contain all
    information needed to display those pages, as well as to navigate between
    them.

    This is not just specific to websites. Apps might, for example, have
    a settings page, a profile page, a help page, and so on.

    Pages are passed directly to the app during construction, like so:

    ```python
    import rio

    app = rio.App(
        build=lambda: rio.Column(
            rio.Text("Welcome to my app!"),
            rio.PageView(grow_y=True),
        ),
        pages=[
            rio.ComponentPage(
                name="Home",
                url_segment="",
                build=lambda: rio.Text("This is the home page"),
            ),
            rio.ComponentPage(
                name="Subpage",
                url_segment="subpage",
                build=lambda: rio.Text("This is a subpage"),
            ),
        ],
    )

    app.run_in_browser()
    ```

    This will display "This is the home page" when navigating to the root URL,
    but "This is a subpage" when navigating to "/subpage". Note that on both
    pages the text "Welcome to my page!" is displayed above the page content.
    That's because it's not part of the `PageView`.

    For additional details, please refer to the how-to guide:
    `https://rio.dev/docs/howto/multiple-pages`.

    ## Attributes

    `name`: A human-readable name for the page. While the page itself doesn't
        use this value directly, it serves as important information for
        debugging, as well as other components such as navigation bars.

    `url_segment`: The URL segment at which this page should be displayed. For
        example, if this is "subpage", then the page will be displayed at
        "https://yourapp.com/subpage". If this is "", then the page will be
        displayed at the root URL.

    `build`: A callback that is called when this page is displayed. It should
        return a Rio component.

    `icon`: The name of an icon to associate with the page. While the page
        itself doesn't use this value directly, it serves as additional
        information for other components such as navigation bars.

    `children`: A list of child pages. These pages will be displayed when
        navigating to a sub-URL of this page. For example, if this page's
        `url_segment` is "page1", and it has a child page with `url_segment`
        "page2", then the child page will be displayed at
        "https://yourapp.com/page1/page2".

    `meta_tags`: A dictionary of meta tags to include in the page's HTML. These
        are used by search engines and social media sites to display
        information about your page.

    `guard`: A callback that is called before this page is displayed. It
        can prevent users from accessing pages which they are not allowed to
        see. For example, you may want to redirect users to your login page
        if they are trying to access their profile page without being
        logged in.

        The callback should return `None` if the user is allowed to access
        the page, or a string or `rio.URL` if the user should be redirected
        to a different page.
    """

    name: str
    url_segment: str
    build: t.Callable[[], rio.Component]
    _: KW_ONLY
    icon: str = DEFAULT_ICON
    children: t.Sequence[ComponentPage | Redirect] = field(default_factory=list)
    guard: t.Callable[[rio.GuardEvent], None | rio.URL | str] | None = None
    meta_tags: dict[str, str] = field(default_factory=dict)

    # This is used to allow users to order pages when using the `rio.page`
    # decorator. It's not public, but simply a convenient place to store this.
    _page_order_: int | None = field(default=None, init=False)

    def __post_init__(self) -> None:
        # In Rio, URLs are case insensitive. An easy way to enforce this, and
        # also prevent casing issues in the user code is to make sure the page's
        # URL fragment is lowercase.
        if self.url_segment != self.url_segment.lower():
            raise ValueError(
                f"Page URLs have to be lowercase, but `{self.url_segment}` is not"
            )

        if "/" in self.url_segment:
            raise ValueError(f"URL segments may not contain slashes")


# Allow using the old `page_url` parameter instead of the new `url_segment`
_old_component_page_init = ComponentPage.__init__


def _new_component_page_init(self, *args, **kwargs) -> None:
    # Rename the parameter
    if "page_url" in kwargs:
        deprecations.warn_parameter_renamed(
            since="0.10",
            old_name="page_url",
            new_name="url_segment",
            owner="rio.ComponentPage",
        )
        kwargs["url_segment"] = kwargs.pop("page_url")

    # Call the original constructor
    _old_component_page_init(self, *args, **kwargs)


ComponentPage.__init__ = _new_component_page_init


@introspection.set_signature(ComponentPage)
def Page(*args, **kwargs):
    deprecations.warn(
        since="0.10",
        message="`rio.Page` has been renamed to `rio.ComponentPage`",
    )
    return ComponentPage(*args, **kwargs)


def _get_active_page_instances(
    available_pages: t.Iterable[rio.ComponentPage | rio.Redirect],
    remaining_segments: tuple[str, ...],
) -> list[rio.ComponentPage | rio.Redirect]:
    """
    Given a list of available pages, and a URL, return the list of pages that
    would be active if navigating to that URL.
    """
    # Get the page responsible for this segment
    try:
        page_segment = remaining_segments[0]
    except IndexError:
        page_segment = ""

    for page in available_pages:
        if page.url_segment == page_segment:
            break
    else:
        return []

    active_pages = [page]

    # Recurse into the children
    if isinstance(page, rio.ComponentPage):
        active_pages += _get_active_page_instances(
            page.children,
            remaining_segments[1:],
        )

    return active_pages


@t.final
@rio.docs.mark_constructor_as_private
@dataclass(frozen=True)
class GuardEvent:
    """
    Holds information regarding a guard event.

    This is a simple dataclass that stores useful information for the guard
    event. They can prevent users from accessing pages which they are not
    allowed to see.

    ## Attributes

    `session`: The current session.

    `active_pages`: All pages that will be active if the navigation succeeds.
    """

    # The current session
    session: rio.Session

    # The pages that would be activated by this navigation
    #
    # This is an `Sequence` rather than `list`, because the same event instance
    # is reused for multiple event handlers. This allows to assign a tuple, thus
    # preventing modifications.
    active_pages: t.Sequence[ComponentPage | Redirect]


def check_page_guards(
    sess: rio.Session,
    target_url_absolute: rio.URL,
) -> tuple[tuple[ComponentPage, ...], rio.URL]:
    """
    Check whether navigation to the given target URL is possible.

    This finds the pages that would be activated by this navigation and runs
    their guards. If the guards effect a redirect, it instead attempts to
    navigate to the redirect target, and so on.

    Raises `NavigationFailed` if navigation to the target URL is not possible
    because of an error, such as an exception in a guard.

    If the URL points to a page which doesn't exist that is not considered an
    error. The result will still be valid. That is because navigation is
    possible, it's just that some PageViews will display a 404 page.

    This function does not perform any actual navigation. It simply checks
    whether navigation to the target page is possible.
    """

    assert target_url_absolute.is_absolute(), target_url_absolute

    # Is any guard opposed to this page?
    initial_target_url = target_url_absolute
    visited_redirects = {target_url_absolute}
    past_redirects = [target_url_absolute]

    while True:
        # TODO: What if the URL is not a child of the base URL? i.e. redirecting
        #   to a completely different site
        target_url_relative = utils.make_url_relative(
            sess._base_url, target_url_absolute
        )

        # Find all pages which would by activated by this navigation
        active_page_instances = tuple(
            _get_active_page_instances(
                sess.app.pages, target_url_relative.parts
            )
        )

        # Check the guards for each activated page
        redirect = None
        event = GuardEvent(
            session=sess,
            active_pages=active_page_instances,
        )

        for page in active_page_instances:
            if isinstance(page, Redirect):
                redirect = page.target
            else:
                if page.guard is None:
                    continue

                try:
                    redirect = page.guard(event)
                except Exception as err:
                    message = f"Rejecting navigation to `{initial_target_url}` because of an error in a page guard of `{page.url_segment}`: {err}"
                    logging.exception(message)
                    raise NavigationFailed(message)

            if redirect is not None:
                break

        # All guards are satisfied - done
        if redirect is None:
            assert all(
                isinstance(page, rio.ComponentPage)
                for page in active_page_instances
            ), active_page_instances

            active_page_instances = t.cast(
                tuple[ComponentPage, ...], active_page_instances
            )

            return active_page_instances, target_url_absolute

        # A guard wants to redirect to a different page
        if isinstance(redirect, str):
            redirect = rio.URL(redirect)

        redirect = sess._active_page_url.join(redirect)

        assert redirect.is_absolute(), redirect

        # Detect infinite loops and break them
        if redirect in visited_redirects:
            page_strings = [
                str(url_segment) for url_segment in past_redirects + [redirect]
            ]
            page_strings_list = "\n -> ".join(page_strings)

            message = f"Rejecting navigation to `{initial_target_url}` because page guards have created an infinite loop:\n\n    {page_strings_list}"
            logging.warning(message)
            raise NavigationFailed(message)

        # Remember that this page has been visited before
        visited_redirects.add(redirect)
        past_redirects.append(redirect)

        # Rinse and repeat
        target_url_absolute = redirect


BuildFunction = t.Callable[[], "rio.Component"]
C = t.TypeVar("C", bound=BuildFunction)


BUILD_FUNCTIONS_FOR_PAGES = dict[BuildFunction, ComponentPage]()


def page(
    *,
    url_segment: str | None = None,
    name: str | None = None,
    icon: str = DEFAULT_ICON,
    guard: t.Callable[[GuardEvent], None | rio.URL | str] | None = None,
    meta_tags: dict[str, str] | None = None,
    order: int | None = None,
):
    """
    This decorator creates a page (complete with URL, icon, etc) that displays
    the decorated component. All parameters are optional, and if omitted,
    sensible defaults will be inferred based on the name of the decorated class.

    In order to create a "root" page, set the `url_segment` to an empty string:

        @rio.page(
            url_segment="",
        )
        class HomePage(rio.Component):
            def build(self):
                return rio.Text(
                    "Welcome to my website",
                    style="heading1",
                )

    For additional details, please refer to the how-to guide [Multiple
    Pages](https://rio.dev/docs/howto/multiple-pages).


    ## Parameters

    `url_segment`: The URL segment at which this page should be displayed. For
        example, if this is "subpage", then the page will be displayed at
        "https://yourapp.com/subpage". If this is "", then the page will be
        displayed at the root URL.

    `name`: A human-readable name for the page. While the page itself doesn't
        use this value directly, it serves as important information for
        debugging, as well as other components such as navigation bars.

    `icon`: The name of an icon to associate with the page. While the page
        itself doesn't use this value directly, it serves as additional
        information for other components such as navigation bars.

    `meta_tags`: A dictionary of meta tags to include in the page's HTML. These
        are used by search engines and social media sites to display
        information about your page.

    `guard`: A callback that is called before this page is displayed. It
        can prevent users from accessing pages which they are not allowed to
        see. For example, you may want to redirect users to your login page
        if they are trying to access their profile page without being
        logged in.

        The callback should return `None` if the user is allowed to access
        the page, or a string or `rio.URL` if the user should be redirected
        to a different page.

    `order`: An int that controls the order of this page relative to its
        siblings. Similar to the `name`, this is relevant for navigation bars.
    """

    def decorator(build: C) -> C:
        nonlocal name, url_segment

        # Derive a default name
        if name is None:
            name = (
                convert_case(build.__name__, "snake").replace("_", " ").title()
            )

        # Derive a default URL segment
        if url_segment is None:
            url_segment = convert_case(build.__name__, "kebab").lower()

        # Create the result
        page = ComponentPage(
            name=name,
            url_segment=url_segment,
            build=build,
            icon=icon,
            guard=guard,
            meta_tags=meta_tags or {},
        )

        # The component page has a field specifically so this decorator can
        # store the page order. However, since this is a frozen dataclass,
        # contortions are needed
        page.__dict__["_page_order_"] = order

        # Store the result
        BUILD_FUNCTIONS_FOR_PAGES[build] = page

        # Return the original class
        return build

    return decorator


def _page_sort_key(page: rio.ComponentPage) -> tuple:
    """
    Returns a key that can be used to sort pages.
    """
    return (
        # Put the home page first
        not page.url_segment == "",
        # Then sort by name
        page.name,
    )


def auto_detect_pages(
    directory: Path,
    *,
    package: str | None = None,
) -> list[rio.ComponentPage]:
    # Find all pages using the iterator method
    pages = list(
        _auto_detect_pages_iter(
            directory,
            package=package,
        )
    )

    # Sort them, ignoring any user-specified ordering for now
    pages.sort(key=_page_sort_key)

    # Now apply the user-specified ordering. This sorting is stable, hence the
    # previous step.
    utils.soft_sort(pages, key=lambda page: page._page_order_)

    # Done
    return pages


def _auto_detect_pages_iter(
    directory: Path,
    *,
    package: str | None = None,
) -> t.Iterable[rio.ComponentPage]:
    try:
        contents = list(directory.iterdir())
    except FileNotFoundError:
        return

    for file_path in contents:
        if not utils.is_python_script(file_path):
            continue

        yield _page_from_python_file(file_path, package)


def _page_from_python_file(
    file_path: Path, package: str | None
) -> ComponentPage:
    module_name = file_path.stem
    if package is not None:
        module_name = package + "." + module_name

    try:
        module = utils.load_module_from_path(file_path, module_name=module_name)
    except BaseException as error:
        # Can't import the module? Display a warning and a placeholder component
        warnings.warn(
            f"Failed to import file '{file_path}': {type(error)} {error}"
        )
        page = _error_page_from_file_name(
            file_path,
            error_summary=f"Failed to import '{file_path}'",
            error_details=f"{type(error)}: {error}",
        )
    else:
        # Search the module for the callable decorated with `@rio.page`
        for obj in vars(module).values():
            try:
                page = BUILD_FUNCTIONS_FOR_PAGES[obj]
                break
            except (TypeError, KeyError):
                continue
        else:
            # Nothing found? Display a warning and a placeholder component
            warnings.warn(
                f"The file {file_path} doesn't seem to contain a page"
                f" definition. Did you forget to decorate your component/build"
                f" function with `@rio.page(...)`?"
            )
            page = _error_page_from_file_name(
                file_path,
                error_summary=f"No page found in '{file_path}'",
                error_details=f"No component in this file was decorated with `@rio.page(...)`",
            )

    # Add sub-pages, if any
    sub_pages = t.cast(list, page.children)
    sub_pages.clear()  # Avoid duplicate entries if this function is called twice
    sub_pages += auto_detect_pages(
        file_path.with_suffix(""),
        package=module_name,
    )

    return page


def _error_page_from_file_name(
    file_path: Path, error_summary: str, error_details: str
) -> ComponentPage:
    return ComponentPage(
        name=convert_case(file_path.stem, "snake").replace("_", " ").title(),
        url_segment=convert_case(file_path.stem, "kebab").lower(),
        build=lambda: rio.components.error_placeholder.ErrorPlaceholder(
            error_summary, error_details
        ),
    )



from __future__ import annotations

import pathlib
import typing as t
from dataclasses import KW_ONLY, dataclass

from uniserde import JsonDoc

import rio

from . import utils
from .color import Color
from .fills import ImageFill, LinearGradientFill, SolidFill
from .self_serializing import SelfSerializing

__all__ = [
    "Font",
    "TextStyle",
]


_TextFill = SolidFill | LinearGradientFill | ImageFill | Color


class UnsetType:
    pass


UNSET = UnsetType()


@dataclass(frozen=True)
class Font(SelfSerializing):
    """
    A custom font face.

    The `Font` class lets you create custom fonts for use in your rio app. To
    instantiate a `Font`, you must pass it at least one font file. (As far as
    rio is concerned, the file format is irrelevant - all that matters is that
    a browser can display it.)


    ## Attributes

    `regular`: The regular (i.e. not bold, not italic) font file.

    `bold`: The bold font file.

    `italic`: The italic font file.

    `bold_italic`: The bold and italic font file.

    `ROBOTO`: A pre-defined font:
        [`Roboto`](https://fonts.google.com/specimen/Roboto).

    `ROBOTO_MONO`: A pre-defined font:
        [`Roboto Mono`](https://fonts.google.com/specimen/Roboto+Mono).
    """

    regular: pathlib.Path | bytes
    bold: pathlib.Path | bytes | None = None
    italic: pathlib.Path | bytes | None = None
    bold_italic: pathlib.Path | bytes | None = None

    def _serialize(self, sess: rio.Session) -> str:
        return sess._register_font(self)

    # Predefined fonts
    ROBOTO: t.ClassVar[Font]
    ROBOTO_MONO: t.ClassVar[Font]


Font.ROBOTO = Font(
    regular=utils.RIO_ASSETS_DIR / "fonts/Roboto/Roboto-Regular.ttf",
    bold=utils.RIO_ASSETS_DIR / "fonts/Roboto/Roboto-Bold.ttf",
    italic=utils.RIO_ASSETS_DIR / "fonts/Roboto/Roboto-Italic.ttf",
    bold_italic=utils.RIO_ASSETS_DIR / "fonts/Roboto/Roboto-BoldItalic.ttf",
)

Font.ROBOTO_MONO = Font(
    regular=utils.RIO_ASSETS_DIR / "fonts/Roboto Mono/RobotoMono-Regular.ttf",
    bold=utils.RIO_ASSETS_DIR / "fonts/Roboto Mono/RobotoMono-Bold.ttf",
    italic=utils.RIO_ASSETS_DIR / "fonts/Roboto Mono/RobotoMono-Italic.ttf",
    bold_italic=utils.RIO_ASSETS_DIR
    / "fonts/Roboto Mono/RobotoMono-BoldItalic.ttf",
)


@dataclass(frozen=True)
class TextStyle(SelfSerializing):
    """
    A collection of styling properties for text.

    Stores styling information for text - font, size, color, etc.


    ## Attributes

    `font`: The `Font` to use for the text. When set to `None`, the default font
        for the current context (heading or regular text, etc) will be used.

    `fill`: The fill (color, gradient, etc.) for the text.

    `font_size`: The font size.

    `italic`: Whether the text is *italic* or not.

    `font_weight`: Whether the text is normal or **bold**.

    `underlined`: Whether the text is underlined or not.

    `strikethrough`: Whether the text should have ~~a line through it~~.

    `all_caps`: Whether the text is transformed to ALL CAPS or not.
    """

    _: KW_ONLY
    font: Font | None = None
    fill: _TextFill | None = None
    font_size: float = 1.0
    italic: bool = False
    font_weight: t.Literal["normal", "bold"] = "normal"
    underlined: bool = False
    strikethrough: bool = False
    all_caps: bool = False

    def replace(
        self,
        *,
        font: Font | None = None,
        fill: _TextFill | None | UnsetType = UNSET,
        font_size: float | None = None,
        italic: bool | None = None,
        font_weight: t.Literal["normal", "bold"] | None = None,
        underlined: bool | None = None,
        strikethrough: bool | None = None,
        all_caps: bool | None = None,
    ) -> TextStyle:
        return type(self)(
            font=self.font if font is None else font,
            fill=self.fill if isinstance(fill, UnsetType) else fill,
            font_size=self.font_size if font_size is None else font_size,
            italic=self.italic if italic is None else italic,
            font_weight=(
                self.font_weight if font_weight is None else font_weight
            ),
            underlined=self.underlined if underlined is None else underlined,
            strikethrough=(
                self.strikethrough if strikethrough is None else strikethrough
            ),
            all_caps=self.all_caps if all_caps is None else all_caps,
        )

    def _serialize(self, sess: rio.Session) -> JsonDoc:
        return {
            "fontName": None
            if self.font is None
            else self.font._serialize(sess),
            "fill": sess._serialize_fill(self.fill),
            "fontSize": self.font_size,
            "italic": self.italic,
            "fontWeight": self.font_weight,
            "underlined": self.underlined,
            "strikethrough": self.strikethrough,
            "allCaps": self.all_caps,
        }



from __future__ import annotations

import typing as t
from dataclasses import KW_ONLY

from uniserde import Jsonable, JsonDoc

from .. import assets
from ..utils import EventHandler, ImageLike
from .fundamental_component import FundamentalComponent

__all__ = ["Image"]


@t.final
class Image(FundamentalComponent):
    """
    Displays a raster image or SVG.

    `Image` does just what you'd expect: it displays a single image. The image
    can be loaded from a URL or a local file.

    The resolution of the image does not affect the size at which it is
    displayed. The `Image` component is flexible with its space requirements and
    adapts to any space allocated by its parent component.

    Note that unlike most components in Rio, the `Image` component does not have
    a `natural` size, since images can be easily be scaled to fit any space.
    Because of this, `Image` defaults to a width and height of 2. This avoids
    invisible images when you forget to set the size.

    The actual picture content can be scaled to fit the assigned shape in one of
    three ways:

    - `"fit"`: The image is scaled to fit entirely inside the shape, while
      maintaining its aspect ratio. This is the default.
    - `"stretch"`: The image is stretched to fill the shape, distorting it if
      necessary.
    - `"zoom"`: The image is scaled to fill the shape entirely, while maintaining
      its aspect ratio. This may cause the image to overflow the shape.

    The image data may be provided in several ways. If the image is already
    hosted somewhere, you can provide a `rio.URL` for easy access. To display a
    local file use a `pathlib.Path` object. Finally, if you already have the
    image data in memory you can either pass it as a `bytes` object or as a
    `PIL.Image` object.


    ## Attributes

    `image`: The image to display.

    `fill_mode`: How the image should be scaled to fit the shape. If `fit`,
        the image is scaled to fit entirely inside the shape. If `stretch`,
        the image is stretched to fill the shape exactly, possibly
        distorting it in the process. If `zoom`, the image is scaled to fill
        the shape entirely, possibly overflowing.

    `on_error`: A function, triggered if the image fails to load.

    `corner_radius`: How round to make the corners of the image. If a single
        number is given, all four corners will be rounded equally. If a
        tuple of four numbers is given, they will be interpreted as the
        radii of the top-left, top-right, bottom-right, and bottom-left
        corners, in that order.


    ## Examples

    This minimal example will display an image hosted on the web:

    ```python
    rio.Image(
        rio.URL("https://example.com/image.png"),
        min_width=20,
        min_height=20,
    )
    ```

    You can also display images from a path. Note that Rio uses modern python
    `pathlib.Path` objects rather than plain strings. The image will be scaled
    to fit the shape, and the corners will be rounded with a radius of 2:

    ```python
    from pathlib import Path

    rio.Image(
        Path("example_image.png"),
        fill_mode="fit",
        min_width=20,
        min_height=20,
        corner_radius=2,
    )
    ```

    You can access the `App`'s assets directory using the `assets` property. This
    will return a `pathlib.Path` object pointing to the assets directory. The
    image will be scaled to fit the shape, and the corners will be rounded with
    a radius of 2:

    ```python
    rio.Image(
        Path(self.session.assets / "example_image.png"),
        fill_mode="fit",
        min_width=20,
        min_height=20,
        corner_radius=2,
    )
    ```
    """

    image: ImageLike
    _: KW_ONLY
    fill_mode: t.Literal["fit", "stretch", "zoom"] = "fit"
    on_error: EventHandler[[]] = None
    corner_radius: float | tuple[float, float, float, float] = 0
    accessibility_description: str = ""

    def __init__(
        self,
        image: ImageLike,
        *,
        fill_mode: t.Literal["fit", "stretch", "zoom"] = "fit",
        on_error: EventHandler[[]] | None = None,
        corner_radius: float | tuple[float, float, float, float] = 0,
        accessibility_description: str = "",
        key: str | int | None = None,
        margin: float | None = None,
        margin_x: float | None = None,
        margin_y: float | None = None,
        margin_left: float | None = None,
        margin_top: float | None = None,
        margin_right: float | None = None,
        margin_bottom: float | None = None,
        min_width: float = 2,
        min_height: float = 2,
        # MAX-SIZE-BRANCH max_width: float | None = None,
        # MAX-SIZE-BRANCH max_height: float | None = None,
        grow_x: bool = False,
        grow_y: bool = False,
        align_x: float | None = None,
        align_y: float | None = None,
        # SCROLLING-REWORK scroll_x: t.Literal["never", "auto", "always"] = "never",
        # SCROLLING-REWORK scroll_y: t.Literal["never", "auto", "always"] = "never",
    ) -> None:
        super().__init__(
            key=key,
            margin=margin,
            margin_x=margin_x,
            margin_y=margin_y,
            margin_left=margin_left,
            margin_top=margin_top,
            margin_right=margin_right,
            margin_bottom=margin_bottom,
            min_width=min_width,
            min_height=min_height,
            # MAX-SIZE-BRANCH max_width=max_width,
            # MAX-SIZE-BRANCH max_height=max_height,
            grow_x=grow_x,
            grow_y=grow_y,
            align_x=align_x,
            align_y=align_y,
            # SCROLLING-REWORK scroll_x=scroll_x,
            # SCROLLING-REWORK scroll_y=scroll_y,
        )

        self.image = image
        self.fill_mode = fill_mode
        self.on_error = on_error
        self.corner_radius = corner_radius
        self.accessibility_description = accessibility_description

    def _get_image_asset(self) -> assets.Asset:
        image = self.image

        if getattr(self, "_image_for_cached_asset", None) != image:
            self._cached_image_asset = assets.Asset.from_image(image)
            self._image_for_cached_asset = image

        return self._cached_image_asset

    def _custom_serialize_(self) -> JsonDoc:
        if isinstance(self.corner_radius, (int, float)):
            corner_radius = (self.corner_radius,) * 4
        else:
            corner_radius = self.corner_radius

        return {
            "imageUrl": self._get_image_asset()._serialize(self.session),
            "reportError": self.on_error is not None,
            "corner_radius": corner_radius,
        }

    async def _on_message_(self, message: Jsonable) -> None:
        await self.call_event_handler(self.on_error)


Image._unique_id_ = "Image-builtin"



from dataclasses import field

# <additional-imports>
from pathlib import Path

import numpy as np
import pandas as pd
from pycoingecko import CoinGeckoAPI  # type: ignore (hidden from user)

import rio

from .. import components as comps
from .. import data_models

# </additional-imports>


# <component>
@rio.page(
    name="Dashboard",
    url_segment="",
)
class DashboardPage(rio.Component):
    """
    A component for displaying our cryptocurrency dashboard.

    The DashboardPage class is the main component of the dashboard application,
    designed to display a user's cryptocurrency balance and the historical data
    of three cryptocurrencies.

    The data is either fetched from the CoinGecko API or loaded from a local CSV
    file.


    ## Attributes

    `coin_data`: the historical data of our portfolio consisting of Bitcoin,
        Litecoin, and Ethereum.
    """

    fetch_data_from_api: bool = False
    coin_data: pd.DataFrame = field(
        default=pd.DataFrame(
            {
                "date": np.zeros(5),
                **{coin: np.zeros(5) for coin in data_models.CRYPTO_LIST},
            }
        )
    )

    @rio.event.on_populate
    async def on_populate(self) -> None:
        """
        The method fetches the historical data of our cryptocurrencies from the
        CoinGecko API or reads the data from a local CSV file, depending on the
        value of the fetch_data_from_api flag.

        The @rio.event.on_populate decorator triggers our on_populate method
        after the component has been created or has been reconciled. This allows
        us to asynchronously fetch any data which depends on the component's
        state.
        """
        if self.fetch_data_from_api is True:
            self.coin_data = self._fetch_coin_data(data_models.CRYPTO_LIST)
        else:
            self.coin_data = self._read_csv(self.session.assets / "cryptos.csv")

    def _read_csv(self, path: Path) -> pd.DataFrame:
        """
        Reads a csv file from the given path and returns a pandas DataFrame.

        ## Parameters

        `path`: A Path object representing the path to the csv file.
        """
        df = pd.read_csv(path)
        df["date"] = pd.to_datetime(df["date"])
        df.set_index("date", inplace=True)
        return df

    def _fetch_coin_data(
        self, coin_names: list[str], vs_currency: str = "usd", days: str = "30"
    ) -> pd.DataFrame:
        """
        This method fetches historical data for a list of cryptocurrencies from the
        CoinGecko API and returns it as a pandas DataFrame.

        The method creates an instance of the CoinGeckoAPI and iterates over the list
        of coin_names. For each coin, it fetches the OHLC (Open, High, Low, Close) data,
        converts it into a DataFrame, and processes it by converting the date to a
        datetime object, setting the date as the index, and dropping the "open", "high",
        and "low" columns. Each processed DataFrame is appended to a list.

        Finally, the method concatenates all the DataFrames in the list into a single
        DataFrame along the columns axis, sets the column names to the coin_names,
        and returns the merged DataFrame.


        ## Parameters

        `coin_names`: A list of strings representing the names of the
            cryptocurrencies to fetch data for.

        `vs_currency`: A string representing the currency to compare
            against. Defaults to "usd".

        `days`: A string representing the number of past days to fetch
            data for. Defaults to "30".
        """

        df_list = []
        cg = CoinGeckoAPI()

        for i in range(len(coin_names)):
            ohlc = cg.get_coin_ohlc_by_id(
                id=coin_names[i], vs_currency=vs_currency, days=days
            )
            df = pd.DataFrame(
                ohlc, columns=["date", "open", "high", "low", "close"]
            )
            df["date"] = pd.to_datetime(df["date"], unit="ms")
            df.set_index("date", inplace=True)
            df = df.drop(columns=["open", "high", "low"])
            df_list.append(df)

        merged_df = pd.concat(df_list, axis=1)
        merged_df.columns = coin_names

        return merged_df

    def build(self) -> rio.Component:
        """
        Returns a grid containing the balance component, three cryptocurrency
        components, and the cryptocurrency chart component.

        The grid is created using the rio.Grid class, which allows us to add
        components to specific rows and columns.

        - The BalanceCard is added to the first row and column. It spans two
            columns.
        - The CryptoChart components are added in the third column to the
            first, second, and third row.
        - The Crypto Chart chart component is added to the second row and
            first column. It spans two rows and spans two columns and rows.

        See the approx. layout below:

        ```
        ╔═══════════════════════════ Grid ═══════════════════════════╗
        ║ ┏━━━━━━━━━ BalanceCard ━━━━━━━━━━┓  ┏━━━━ CryptoCard ━━━━┓ ║
        ║ ┃ custom component               ┃  ┃ c. component (BTC) ┃ ║
        ║ ┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┛  ┗━━━━━━━━━━━━━━━━━━━━┛ ║
        ║ ┏━━━━━━━━━ CryptoChart ━━━━━━━━━━┓  ┏━━━━ CryptoCard ━━━━┓ ║
        ║ ┃ custom component               ┃  ┃ c. component (LTC) ┃ ║
        ║ ┃                                ┃  ┗━━━━━━━━━━━━━━━━━━━━┛ ║
        ║ ┃                                ┃  ┏━━━━ CryptoCard ━━━━┓ ║
        ║ ┃                                ┃  ┃ c. component (ETH) ┃ ║
        ║ ┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┛  ┗━━━━━━━━━━━━━━━━━━━━┛ ║
        ╚════════════════════════════════════════════════════════════╝
        ```
        """

        grid = rio.Grid(
            column_spacing=2,
            row_spacing=2,
            align_y=0.3,
            margin_left=5,
            margin_right=5,
        )

        grid.add(
            comps.BalanceCard(data=self.coin_data),
            row=0,
            column=0,
        )

        # you can use a loop to add the CryptoCard components for each coin
        for i, coin in enumerate(data_models.CRYPTO_LIST):
            grid.add(
                comps.CryptoCard(
                    data=self.coin_data,
                    coin=coin,
                    coin_amount=data_models.MY_COINS[coin][0],
                    coin_ticker=data_models.MY_COINS[coin][1],
                    color=data_models.MY_COINS[coin][2],
                    logo_url=data_models.MY_COINS[coin][3],
                ),
                row=i,
                column=1,
            )

        grid.add(
            comps.CryptoChart(data=self.coin_data, coin="bitcoin"),
            row=1,
            column=0,
            height=2,
        )

        return grid


# </component>



import subprocess
import sys

import revel


def main() -> None:
    revel.print("Running sanity checks...")

    ensure_branch("main")
    ensure_no_uncommitted_changes()
    ensure_up_to_date_with_remote()

    build_frontend()
    ensure_tests_pass()

    revel.print("Everything is in order.")
    make_new_release()


def ensure_branch(branch_name: str) -> None:
    process = subprocess.run(
        ["git", "branch", "--show-current"],
        check=True,
        capture_output=True,
        text=True,
    )
    if process.stdout.strip("\n") != branch_name:
        revel.fatal(f'You must checkout the "{branch_name}" branch')


def ensure_no_uncommitted_changes() -> None:
    process = subprocess.run(
        ["git", "status", "--porcelain", "--untracked-files=no"],
        check=True,
        capture_output=True,
        text=True,
    )
    if process.stdout.strip():
        revel.fatal("There are uncommitted changes")


def ensure_up_to_date_with_remote() -> None:
    subprocess.run(["git", "remote", "update"], check=True)
    process = subprocess.run(
        ["git", "status", "--porcelain", "--untracked-files=no"],
        check=True,
        capture_output=True,
        text=True,
    )
    if process.stdout.strip():
        revel.fatal("Local branch is not up-to-date with remote")


def build_frontend() -> None:
    subprocess.run(
        ["rye", "run", "build"],
        check=True,
    )


def ensure_tests_pass() -> None:
    process = subprocess.run(["rye", "test", "--", "-x", "--disable-warnings"])

    if process.returncode == 0:
        revel.print("Everything is in order. Publishing...")
    else:
        if not revel.select_yes_no(
            "The test suite does not pass. Do you want to publish anyway?"
        ):
            sys.exit(1)

        revel.print("Publishing...")


def make_new_release() -> None:
    # Bump version
    bump_version()
    subprocess.run(["git", "push"], check=True)

    # Create a tag
    version = get_version()
    subprocess.run(["git", "tag", version], check=True)
    subprocess.run(["git", "push", "origin", "tag", version], check=True)

    # Publish
    subprocess.run(["rye", "build", "--clean"], check=True)
    subprocess.run(["rye", "publish"], check=True)


def bump_version() -> None:
    options = {
        "patch": "patch",
        "minor": "minor",
        "major": "major",
    }

    version = get_version()
    if "rc" in version:
        options["release candidate"] = "rc"

    version_part_to_bump = revel.select(
        options,
        prompt="Which version do you want to bump?",
    )

    if version_part_to_bump != "rc":
        if revel.select_yes_no("Make this a pre-release?"):
            version_part_to_bump += ",rc"

    subprocess.run(
        [sys.executable, "-m", "hatch", "version", version_part_to_bump],
        check=True,
    )
    subprocess.run(["git", "commit", "--all", "-m", "bump version"], check=True)


def get_version() -> str:
    return subprocess.run(
        [sys.executable, "-m", "hatch", "version"],
        check=True,
        capture_output=True,
        text=True,
    ).stdout.strip()


if __name__ == "__main__":
    main()



from __future__ import annotations

import hashlib
import os
import secrets
import uuid
from dataclasses import dataclass
from datetime import datetime, timezone

import rio


@dataclass
class UserSettings(rio.UserSettings):
    """
    Model for data stored client-side for each user.
    """

    # The (possibly expired) authentication token for this user. If this matches
    # the id of any valid `UserSession`, it is safe to consider the user as
    # authenticated being in that session.
    #
    # This prevents users from having to log-in again each time the page is
    # accessed.
    auth_token: str


@dataclass
class UserSession:
    # This ID uniquely identifies the session. It also serves as the
    # authentication token for the user.
    id: str

    # The user this session belongs to
    user_id: uuid.UUID

    # When this session was initially created
    created_at: datetime

    # Until when this session is valid
    valid_until: datetime


@dataclass
class AppUser:
    """
    Model for a user of the application.
    """

    # A unique identifier for this user
    id: uuid.UUID

    # The user's chosen username
    username: str

    # When the user account was created
    created_at: datetime

    # The hash and salt of the user's password. By storing these values we can
    # verify that a user entered the correct password without storing the actual
    # password in the database. Google "hashing & salting" for details if you're
    # curious.
    password_hash: bytes
    password_salt: bytes

    @classmethod
    def new_with_defaults(cls, username, password) -> AppUser:
        """
        Create a new user with the given username and password, filling in
        reasonable defaults for the other fields.
        """

        password_salt = os.urandom(64)

        return AppUser(
            id=uuid.uuid4(),
            username=username,
            created_at=datetime.now(timezone.utc),
            password_hash=cls.get_password_hash(password, password_salt),
            password_salt=password_salt,
        )

    @classmethod
    def get_password_hash(cls, password, password_salt: bytes) -> bytes:
        """
        Compute the hash of a password using a given salt.
        """
        return hashlib.pbkdf2_hmac(
            "sha256",
            password.encode("utf-8"),
            password_salt,
            100000,
        )

    def password_equals(self, password: str) -> bool:
        """
        Safely compare a password to the stored hash. This differs slightly from
        the `==` operator in that it is resistant to timing attacks.
        """
        return secrets.compare_digest(
            self.password_hash,
            self.get_password_hash(password, self.password_salt),
        )



import typing as t

import rio.debug.dev_tools.component_picker
import rio.debug.dev_tools.component_tree

from . import component_attributes, layout_subpage

MARGIN = 1


class TreePage(rio.Component):
    """
    This page displays all components in the app as an interactive tree.
    Components can be selected to view their details. This comes with a fair
    amount of complexity, since the page effectively switches between multiple
    views.
    """

    current_view: t.Literal["tree", "attributes", "layout"] = "tree"

    # This can be invalid. The component must deal with it.
    selected_component_id: int = -1

    def _switch_to_tree(self, _: rio.PressEvent) -> None:
        self.current_view = "tree"

    def _switch_to_details(self) -> None:
        self.current_view = "attributes"

    def _switch_to_layout(self) -> None:
        self.current_view = "layout"

    def _build_back_menu(self, label: str) -> rio.Component:
        return rio.MouseEventListener(
            rio.Rectangle(
                content=rio.Row(
                    rio.Icon(
                        "material/arrow_back",
                        min_width=2.2,
                        min_height=2.2,
                    ),
                    rio.Text(
                        label,
                        selectable=False,
                        style="heading2",
                        justify="left",
                        grow_x=True,
                    ),
                    spacing=1,
                    margin=MARGIN,
                ),
                fill=rio.Color.TRANSPARENT,
                hover_fill=self.session.theme.neutral_palette.background_active,
                ripple=True,
                cursor=rio.CursorStyle.POINTER,
                transition_time=0.1,
            ),
            on_press=self._switch_to_tree,
        )

    def _build_total_components(self) -> rio.Component:
        """
        Creates a component which displays the total number of components
        """
        n_total_components = len(self.session._weak_components_by_id)
        text = f"{n_total_components} components in total"

        if n_total_components <= 150:
            text += " (fast)"
            icon = "material/speed:fill"
            color = self.session.theme.success_color
        elif n_total_components <= 400:
            text += " (borderline)"
            icon = "material/warning:fill"
            color = self.session.theme.neutral_palette.foreground
        else:
            text += " (slow)"
            icon = "material/warning:fill"
            color = self.session.theme.warning_color

        # (The plural form "components" is safe here, since this page alone
        # has more than one component)
        return rio.Row(
            rio.Icon(
                icon=icon,
                fill=color,
            ),
            rio.Text(
                text,
                style=rio.TextStyle(
                    italic=True,
                    fill=color,
                ),
            ),
            spacing=0.5,
            align_x=0.5,
        )

    def _build_tree_view(self) -> rio.Component:
        return rio.Column(
            rio.Row(
                rio.Text(
                    "Component Tree",
                    style="heading2",
                    justify="left",
                    grow_x=True,
                ),
                rio.Tooltip(
                    rio.debug.dev_tools.component_picker.ComponentPicker(),
                    "Select a component by clicking on it",
                ),
                margin_right=MARGIN,
            ),
            rio.debug.dev_tools.component_tree.ComponentTree(
                component_id=self.bind().selected_component_id,
                grow_y=True,
                # Note how there is no `margin_right` here. This is intentional
                # because the tree has an internal scroll bar which would look
                # silly if floating in space.
            ),
            # self._build_total_components(),
            rio.Button(
                "Attributes",
                icon="material/info",
                shape="rounded",
                on_press=self._switch_to_details,
                margin_right=MARGIN,
            ),
            rio.Button(
                "Layout",
                icon="material/space_dashboard",
                shape="rounded",
                on_press=self._switch_to_layout,
                margin_right=MARGIN,
            ),
            spacing=0.5,
            margin_left=MARGIN,
            margin_top=MARGIN,
            margin_bottom=MARGIN,
        )

    def _build_details_view(self) -> rio.Component:
        assert self.current_view == "attributes", self.current_view
        assert self.selected_component_id is not None

        # Get the target component
        try:
            target = self.session._weak_components_by_id[
                self.selected_component_id
            ]
            heading = type(target).__name__
        except KeyError:
            heading = "Attributes"

        # Delegate the hard work to another component
        return rio.Column(
            self._build_back_menu(heading),
            rio.ScrollContainer(
                component_attributes.ComponentAttributes(
                    component_id=self.bind().selected_component_id,
                    on_switch_to_layout_view=self._switch_to_layout,
                    margin_left=MARGIN,
                    margin_right=MARGIN,
                    margin_bottom=MARGIN,
                ),
                grow_y=True,
                scroll_x="never",
            ),
        )

    def _build_layout_view(self) -> rio.Component:
        assert self.current_view == "layout", self.current_view
        assert self.selected_component_id is not None

        return rio.Column(
            self._build_back_menu("Layout"),
            rio.ScrollContainer(
                layout_subpage.LayoutSubpage(
                    component_id=self.bind().selected_component_id,
                    margin=MARGIN,
                ),
                grow_y=True,
                scroll_x="never",
            ),
        )

    def build(self) -> rio.Component:
        if self.current_view == "tree":
            return self._build_tree_view()

        if self.current_view == "attributes":
            return self._build_details_view()

        if self.current_view == "layout":
            return self._build_layout_view()

        raise AssertionError("Unreachable")



from __future__ import annotations

import typing as t

import rio

from .. import utils
from .component import Component

__all__ = [
    "DialogContainer",
]


@t.final
class DialogContainer(Component):
    build_content: t.Callable[[], Component]
    owning_component_id: int
    is_modal: bool
    is_user_closeable: bool
    on_close: rio.EventHandler[[]]

    def build(self) -> Component:
        return utils.safe_build(self.build_content)

    # Note that this is NOT `_custom_serialize`. Dialog containers are
    # high-level on the Python side, but sent to the client as though they were
    # fundamental. To prevent a whole bunch of custom code in the serializer,
    # this function handles the serialization of dialog containers.
    def serialize(self) -> dict[str, t.Any]:
        return {
            "owning_component_id": self.owning_component_id,
            "is_modal": self.is_modal,
            "is_user_closable": self.is_user_closeable,
        }



from __future__ import annotations

import typing as t

import rio

from .fundamental_component import FundamentalComponent

__all__ = [
    "ClassContainer",
]


class ClassContainer(FundamentalComponent):
    """
    A Component which holds a single child.

    Component which holds a single child, and applies a list of CSS classes to
    it. This is enough to implement several components, preventing the need to
    create a whole bunch of almost identical JavaScript components.

    This component is only intended for internal use and is not part of the
    public API.


    ## Attributes

    `content`: The child component to apply the classes to.

    `classes`: The list of classes to apply to the child component.


    ## Metadata

    `public`: False
    """

    content: rio.Component | None
    classes: t.Sequence[str]

    def _get_debug_details_(self) -> dict[str, t.Any]:
        result = super()._get_debug_details_()
        result.pop("classes")
        return result


ClassContainer._unique_id_ = "ClassContainer-builtin"



from __future__ import annotations

import typing as t
from dataclasses import dataclass

from uniserde import JsonDoc

from .. import maybes
from .fundamental_component import FundamentalComponent

if t.TYPE_CHECKING:
    import numpy  # type: ignore
    import pandas  # type: ignore
    import polars  # type: ignore


__all__ = ["Table"]


TableValue = int | float | str


@t.final
@dataclass
class TableSelection:
    _left: int
    _top: int | t.Literal["header"]
    _width: int
    _height: int

    _font_weight: t.Literal["normal", "bold"] | None = None

    def style(
        self,
        *,
        font_weight: t.Literal["normal", "bold"] | None = None,
    ) -> None:
        if font_weight is not None:
            self._font_weight = font_weight

    def _as_json(self) -> JsonDoc:
        # Some values are always present
        result: JsonDoc = {
            "left": self._left,
            "top": self._top,
            "width": self._width,
            "height": self._height,
        }

        # Styling only if set
        if self._font_weight is not None:
            result["fontWeight"] = self._font_weight

        # Done
        return result


def _index_to_start_and_extent(
    index: int | slice | str,
    size_in_axis: int,
    axis: t.Literal["x", "y"],
) -> tuple[
    int | t.Literal["header"],
    int,
]:
    """
    Given a one-axis `__getitem__` index, returns the start and extent of
    the slice as non-negative integers.

    This function is intentionally a separate function instead of a method
    to allow for easy unit-testing.
    """
    # Python considers booleans to be integers. Guard against them being used as
    # numbers here
    if isinstance(index, bool):
        raise ValueError(
            f"Table indices should be integers or slices, not {index!r}"
        )

    # Integers are easy
    if isinstance(index, int):
        start = index
        stop = index + 1

    # Slices need more work
    elif isinstance(index, slice):
        if index.start is None:
            start = 0
        elif isinstance(index.start, int) and not isinstance(index.start, bool):
            start = index.start
        else:
            raise ValueError(
                f"Table indices should be integers or slices, not {index!r}"
            )

        if index.stop is None:
            stop = size_in_axis
        elif isinstance(index.stop, int) and not isinstance(index.stop, bool):
            stop = index.stop
        else:
            raise ValueError(
                f"Table indices should be integers or slices, not {index!r}"
            )

    # In the y-axis, "header" is a valid index
    elif axis == "y" and index == "header":
        return "header", 1

    # Anything else is invalid
    else:
        raise ValueError(
            f"Table indices should be integers or slices, not {index!r}"
        )

    # Negative numbers count backwards from the end
    if start < 0:
        if start + size_in_axis < 0:
            raise IndexError(
                f"Negative index {start} out of bounds for axis of size {size_in_axis}"
            )

        start = size_in_axis + start

    if stop < 0:
        if stop + size_in_axis < 0:
            raise IndexError(
                f"Negative index {stop} out of bounds for axis of size {size_in_axis}"
            )

        stop = size_in_axis + stop

    # Done
    return start, stop - start


def _string_index_to_start_and_extent(
    index: str | int | slice,
    column_names: list[str] | None,
    size_in_axis: int,
    axis: t.Literal["x", "y"],
) -> tuple[
    int | t.Literal["header"],
    int,
]:
    """
    Same as `_index_to_start_and_extent`, but with support for string indices.
    """
    # If passed a string, select the entire column
    if isinstance(index, str):
        if column_names is None:
            raise ValueError(
                "Cannot index into this table using a column name, since it doesn't have any headers"
            )

        try:
            left = column_names.index(index)
        except ValueError:
            raise KeyError(f"This table doesn't have a column named {index!r}")

        return left, 1

    # Otherwise delegate to the other function
    return _index_to_start_and_extent(
        index,
        size_in_axis,
        axis,
    )


def _indices_to_rectangle(
    index: str | tuple[int | slice | str, str | int | slice],
    column_names: list[str] | None,
    data_width: int,
    data_height: int,
) -> tuple[
    int,
    int | t.Literal["header"],
    int,
    int,
]:
    # Get the raw x & y indices
    if isinstance(index, str):
        index_y = slice(None)
        index_x = index
    else:
        if not isinstance(index, tuple):
            raise ValueError(
                f"Table indices should be a tuple of two elements, not {index!r}"
            )
        elif len(index) != 2:
            raise ValueError(
                f"Table indices should have exactly two elements, not {len(index)}"
            )

        index_y, index_x = index

    # Get the index as a tuple (top, left, height, width)
    top, height = _index_to_start_and_extent(
        index_y,
        data_height,
        axis="y",
    )

    left, width = _string_index_to_start_and_extent(
        index_x,
        column_names,
        data_width,
        axis="x",
    )

    assert isinstance(left, int), left  # Left can't be "header"

    return left, top, width, height


# TODO: add more content to docstring
@t.final
class Table(FundamentalComponent):
    """
    Display & input for tabular data.

    Tables are a way to display data in a grid, with rows and columns. They are
    very useful for displaying data that is naturally tabular, such as
    spreadsheets, databases, or CSV files. Tables can be sorted by clicking on
    the column headers.


    ## Attributes

    `data`: The data to display.

    `show_row_numbers`: Whether to show row numbers on the left side of the table.


    ## Examples

    A simple table with some data:

    ```python
    rio.Table(
        data={
            "Name": ["Alice", "Bob", "Charlie"],
            "Age": [25, 30, 35],
            "City": ["New York", "San Francisco", "Los Angeles"],
        }
    )
    ```

    Indexing into a table to apply styling works similar as numpy arrays.
    The following example makes the second and third row, and the second
    and third column bold:

    ```python
    class MyComponent(rio.Component):
        def build(self) -> rio.Component:
            table = rio.Table(
                data={
                    "Name": ["Alice", "Bob", "Charlie"],
                    "Age": [25, 30, 35],
                    "City": ["New York", "San Francisco", "Los Angeles"],
                }
            )

            # apply some styling to the table, indexing rows and columns
            # works similar as numpy arrays.
            # second and third row, second and third column are bold
            table[1:3, 1:3].style(font_weight="bold")

            return table


    ## Metadata

    `experimental`: True
    """

    data: (
        pandas.DataFrame
        | polars.DataFrame
        | t.Mapping[str, t.Iterable[TableValue]]
        | t.Iterable[t.Iterable[TableValue]]
        | numpy.ndarray
    )
    show_row_numbers: bool = True

    # All headers, if present
    _headers: list[str] | None = None

    # The data, as a list of rows ("row-major"). This is initialized in
    # `__post_init__`.
    _data: list[list[TableValue]] = []

    # All styles applied to the table, in the order they were added.
    _styling: list[TableSelection] = []

    def __post_init__(self) -> None:
        # Bring the data into a standardized format: Either a dataframe (pandas
        # or polars) or a numpy array.
        #
        # Pandas
        if isinstance(self.data, maybes.PANDAS_DATAFRAME_TYPES):
            self._headers = list(self.data.columns)
            self._data = self.data.values.tolist()  # type: ignore

        # Polars
        elif isinstance(self.data, maybes.POLARS_DATAFRAME_TYPES):
            self._headers = list(self.data.columns)
            self._data = self.data.to_numpy().tolist()  # type: ignore

        # Numpy
        elif isinstance(self.data, maybes.NUMPY_ARRAY_TYPES):
            self._headers = None
            self._data = self.data.tolist()

        # Mapping
        elif isinstance(self.data, t.Mapping):
            data = t.cast(t.Mapping[str, t.Iterable[TableValue]], self.data)
            self._headers = list(data.keys())

            # Verify all columns have the same length
            columns: list[list[TableValue]] = []
            column_lengths = set()

            for column in data.values():
                column = list(column)
                column_lengths.add(len(column))
                columns.append(column)

            if len(column_lengths) > 1:
                raise ValueError("All table columns must have the same length")

            # Black magic to transpose the data
            self._data = list(map(list, zip(*self.data.values())))

        # Iterable of iterables
        else:
            data = t.cast(t.Iterable[t.Iterable[TableValue]], self.data)
            self._headers = None
            self._data = []
            row_lengths = set()

            for row in data:
                row = list(row)
                row_lengths.add(len(row))
                self._data.append(row)

            if len(row_lengths) > 1:
                raise ValueError("All table rows must have the same length")

    def _custom_serialize_(self) -> JsonDoc:
        return {
            "headers": self._headers,
            "data": self._data,
            "styling": [style._as_json() for style in self._styling],
        }  # type: ignore

    def _shape(self) -> tuple[int, int]:
        """
        Returns the (height, width) of the table data. This does not include
        headers! This is like numpy's shape but takes into account the many
        different types of data that can be passed to the data attribute.
        """
        try:
            return (len(self._data), len(self._data[0]))
        except IndexError:
            return (0, 0)

    def _column_name_to_int(self, column_name: str) -> int:
        if self._headers is None:
            raise ValueError(
                "Cannot index into this table using a column name, since it doesn't have any headers"
            )

        try:
            return self._headers.index(column_name)
        except ValueError:
            raise KeyError(
                f"This table doesn't have a column named {column_name!r}"
            )

    def __getitem__(
        self,
        index: (
            str
            | tuple[
                int | slice | str,
                int | slice | str,
            ]
        ),
    ) -> TableSelection:
        # Get the index as a tuple (top, left, height, width)
        data_height, data_width = self._shape()

        left, top, width, height = _indices_to_rectangle(
            index,
            self._headers,
            data_width,
            data_height,
        )

        # Verify the indices are within bounds
        right = left + width
        out_of_bounds = (left < 0 or left >= data_width) or (
            right < 0 or right > data_width
        )

        if top == "header":
            assert height == 1, height
        else:
            bottom = top + height
            out_of_bounds = (
                out_of_bounds
                or (top < 0 or top >= data_height)
                or (bottom < 0 or bottom > data_height)
            )

        if out_of_bounds:
            raise IndexError(
                f"The table index {index!r} is out of bounds for a table of size {data_height}x{data_width}"
            )

        # Construct the result
        result = TableSelection(
            _left=left,
            _top=top,
            _width=width,
            _height=height,
            _font_weight=None,
        )

        # Keep track of it
        self._styling.append(result)

        # Done!
        return result


Table._unique_id_ = "Table-builtin"



import functools
import random
import typing as t

import rio

from ... import icon_registry

ICON_SET = "material"
GRID_N_ROWS = 12
GRID_N_COLUMNS = 6


# Choose icons to display
def find_icons_to_display() -> t.Iterable[str]:
    # Get a list of all available icons
    names_and_variants: list[tuple[str, str | None]] = list(
        icon_registry.all_icons_in_set(ICON_SET)
    )

    # Choose some at random
    result = random.sample(names_and_variants, GRID_N_ROWS * GRID_N_COLUMNS)

    # Convert to strings
    for name, variant in result:
        yield (
            f"{ICON_SET}/{name}"
            if variant is None
            else f"{ICON_SET}/{name}:{variant}"
        )


DISPLAYED_ICON_NAMES: list[str] = list(find_icons_to_display())


class SampleIconsGrid(rio.Component):
    """
    Displays a raster of icons.
    """

    on_select_icon: rio.EventHandler[str] = None

    async def _on_select_icon(self, icon_name: str) -> None:
        """
        Called when an icon is selected.
        """
        await self.call_event_handler(self.on_select_icon, icon_name)

    async def _on_randomize(self) -> None:
        global DISPLAYED_ICON_NAMES
        DISPLAYED_ICON_NAMES = list(find_icons_to_display())

        await self.force_refresh()

    def build(self) -> rio.Component:
        # Build a flat list of all icons
        icons_flat: list[rio.Component] = [
            rio.IconButton(
                icon=icon,
                min_size=3,
                style="plain-text",
                on_press=functools.partial(self._on_select_icon, icon),
                key=icon,
                align_x=0.5,
                align_y=0.5,
            )
            for icon in DISPLAYED_ICON_NAMES
        ]

        # Split the flat list into rows
        icon_rows: list[list[rio.Component]] = [
            icons_flat[i : i + GRID_N_COLUMNS]
            for i in range(0, len(icons_flat), GRID_N_COLUMNS)
        ]

        # Build the resulting grid
        return rio.Grid(
            *icon_rows,
            rio.Button(
                "Randomize",
                icon="material/refresh",
                style="colored-text",
                on_press=self._on_randomize,
                align_y=0.5,
            ),
            row_spacing=0.5,
            column_spacing=0.5,
        )



from __future__ import annotations

import typing as t
from dataclasses import KW_ONLY, dataclass

from uniserde import Jsonable

import rio

from . import color, deprecations
from . import text_style as text_style_module

__all__ = [
    "Palette",
    "Theme",
]


T = t.TypeVar("T")


def _derive_color(
    color: rio.Color,
    offset: float,
    *,
    bias_to_bright: float = 0,
    target_color: rio.Color | None = None,
) -> rio.Color:
    # If a target color was provided, move towards that color
    #
    # The more different the two colors are, the less the new color will be
    # used. This helps e.g. with dark themes, which look bonkers if the target
    # color is very light.
    if target_color is not None:
        difference = (
            abs(target_color.red - color.red)
            + abs(target_color.green - color.green)
            + abs(target_color.blue - color.blue)
        )

        if difference < 0.01:
            offset_scale = 1
        else:
            offset_scale = min(1.5 / difference, 1)

        result = color.blend(target_color, offset * offset_scale)

    # Otherwise change the color's brightness
    else:
        threshold = 0.5 + 0.5 * bias_to_bright
        perceived_brightness = color.perceived_brightness
        brighten = perceived_brightness <= threshold

        # Calculate the new color
        if brighten:
            result = color.brighter(offset)
        else:
            result = color.darker(offset)

    # Desaturate the color slightly
    hue, saturation, value = result.hsv
    saturation = max(saturation * 0.9, 0)

    return rio.Color.from_hsv(
        hue=hue,
        saturation=saturation,
        value=value,
        opacity=result.opacity,
    )


def _make_semantic_palette(color: rio.Color) -> Palette:
    return Palette(
        background=color,
        background_variant=_derive_color(
            color,
            0.08,
            bias_to_bright=-0.4,
        ),
        background_active=_derive_color(
            color,
            0.15,
            bias_to_bright=0.8,
        ),
        foreground=_derive_color(
            color,
            0.4,
        ),
    )


@t.final
@dataclass()
class Palette:
    background: rio.Color
    background_variant: rio.Color
    background_active: rio.Color

    foreground: rio.Color

    @staticmethod
    def from_color(
        color: rio.Color,
        *,
        offset: float = 1,
        target_color: rio.Color | None = None,
    ) -> Palette:
        """
        ## Metadata

        experimental: True
        """
        return Palette(
            background=color,
            background_variant=_derive_color(
                color,
                offset=0.1 * offset,
                bias_to_bright=-0.3,
                target_color=target_color,
            ),
            background_active=_derive_color(
                color,
                offset=0.2 * offset,
                bias_to_bright=-0.3,
                target_color=target_color,
            ),
            foreground=_derive_color(
                color,
                offset=0.8,
            ),
        )


@t.final
@dataclass()
class Theme:
    """
    Defines the visual style of the application.

    The `Theme` contains all colors, text styles, and other visual properties
    that are used throughout the application. If you wish to change the
    appearance of your app, this is the place to do it.

    Warning: The exact attributes available in themes are still subject to
        change. The recommended way to create themes is using either the
        `from_colors` or `pair_from_colors` method, as they provide a more
        stable interface.
    """

    _: KW_ONLY

    primary_palette: Palette
    secondary_palette: Palette

    background_palette: Palette
    neutral_palette: Palette
    hud_palette: Palette
    disabled_palette: Palette

    success_palette: Palette
    warning_palette: Palette
    danger_palette: Palette

    # Other
    corner_radius_small: float
    corner_radius_medium: float
    corner_radius_large: float

    shadow_color: rio.Color

    monospace_font: text_style_module.Font

    # Text styles
    heading1_style: rio.TextStyle
    heading2_style: rio.TextStyle
    heading3_style: rio.TextStyle
    text_style: rio.TextStyle

    def __init__(self) -> None:
        # Themes are still very much in flux. New attributes will be added,
        # removed and changed. Stop anyone from trying to create a theme
        # manually, as that is an endeavor doomed to fail.
        """
        Do not use. Themes are still in flux and should be created using the
        `from_colors` or `pair_from_colors` methods.

        ## Metadata

        `public`: False
        """

    @staticmethod
    def _create_new(
        primary_palette: Palette,
        secondary_palette: Palette,
        background_palette: Palette,
        neutral_palette: Palette,
        hud_palette: Palette,
        disabled_palette: Palette,
        success_palette: Palette,
        warning_palette: Palette,
        danger_palette: Palette,
        corner_radius_small: float,
        corner_radius_medium: float,
        corner_radius_large: float,
        shadow_color: rio.Color,
        monospace_font: text_style_module.Font,
        heading1_style: rio.TextStyle,
        heading2_style: rio.TextStyle,
        heading3_style: rio.TextStyle,
        text_style: rio.TextStyle,
    ) -> Theme:
        """
        The Theme's constructor is blocked to prevent users from creating themes
        manually. This is an internal method that bypasses that block.
        """
        self = object.__new__(Theme)

        self.__dict__.update(
            {
                "primary_palette": primary_palette,
                "secondary_palette": secondary_palette,
                "background_palette": background_palette,
                "neutral_palette": neutral_palette,
                "hud_palette": hud_palette,
                "disabled_palette": disabled_palette,
                "success_palette": success_palette,
                "warning_palette": warning_palette,
                "danger_palette": danger_palette,
                "corner_radius_small": corner_radius_small,
                "corner_radius_medium": corner_radius_medium,
                "corner_radius_large": corner_radius_large,
                "shadow_color": shadow_color,
                "monospace_font": monospace_font,
                "heading1_style": heading1_style,
                "heading2_style": heading2_style,
                "heading3_style": heading3_style,
                "text_style": text_style,
            }
        )

        return self

    @classmethod
    @deprecations.parameters_remapped(
        since="0.9.0",
        light=lambda light: {
            "mode": "light" if light else "dark",
        },
    )
    def from_colors(
        cls,
        *,
        primary_color: rio.Color | None = None,
        secondary_color: rio.Color | None = None,
        background_color: rio.Color | None = None,
        neutral_color: rio.Color | None = None,
        hud_color: rio.Color | None = None,
        disabled_color: rio.Color | None = None,
        success_color: rio.Color | None = None,
        warning_color: rio.Color | None = None,
        danger_color: rio.Color | None = None,
        corner_radius_small: float = 0.4,
        corner_radius_medium: float = 0.8,
        corner_radius_large: float = 1.8,
        heading_fill: t.Literal["primary", "plain", "auto"]
        | text_style_module._TextFill = "auto",
        text_color: rio.Color | None = None,
        font: text_style_module.Font = text_style_module.Font.ROBOTO,
        monospace_font: text_style_module.Font = text_style_module.Font.ROBOTO_MONO,
        mode: t.Literal["light", "dark"] = "light",
    ) -> Theme:
        """
        Creates a new theme based on the provided colors.

        Themes store a large number of colors and text styles that are used
        throughout the app. This function provides a convenient way to only
        specify the most important colors of your theme, and have Rio handle the
        rest. It is the recommended way to create a new theme.

        To use a custom theme in your app first create a theme, and then pass it
        to your app instance

        ```python
        # Create a new theme
        theme = rio.Theme.from_colors(
            # Configure your theme here
        )

        # And apply it to your app
        app = rio.App(
            theme=theme,
            # ...
        )
        ```

        Alternatively, there is also `pair_from_colors`, which creates both a
        light and a dark theme at the same time. This is useful if you want Rio
        to automatically switch between the two based on the user's system
        preferences.


        ## Parameters

        `primary_color`: The main color of your app. This color will be used to
            tint the background and by some large components to fill large
            spaces with color.

        `secondary_color`: A color that nicely complements the primary color. It
            is often used by small components such as buttons and switches.

        `background_color`: The app's background color. This should be a neutral
            color that doesn't distract from the content.

        `neutral_color`: Similar to the background color, it is also used for
            neutral areas. It should however be slightly different, allowing you
            to create a visual hierarchy. This is the default color of large
            elements such as cards.

        `hud_color`: Used for elements that pop over the content, such as
            tooltips.

        `disabled_color`: Used by insensitive components to indicate that they
            are not interactive. Typically a shade of grey.

        `success_color`: A color to give positive feedback the user. Typically
            a shade of green.

        `warning_color`: A color to indicate that something might be wrong, but
            isn't critical. Typically orange.

        `danger_color`: A color to indicate that something is wrong and needs
            immediate attention. Typically a shade of red.

        `corner_radius_small`: The corner radius of small components such as
            text inputs

        `corner_radius_medium`: The corner radius of medium-sized components,
            such as small cards.

        `corner_radius_large`: The corner radius of large components, such as
            large cards and dialogs.

        `heading_fill`: The fill to use for headings. This allows you to specify
            a more interesting color, or even a gradient. If set to `"auto"`,
            Rio will automatically switch between the primary color and a plain
            text color based on legibility.

            This only affects headings in background and neutral contexts.

        `text_color`: The default text color to use for regular text. Please
            note that this only applies to text in a neutral or background
            context. Text that's e.g. placed on a `rio.Card` with
            `color="primary"` will use a different color to ensure legibility.

        `font`: The default font to use when no other is specified.

        `monospace_font`: The font to use for monospace text, such as code.

        `mode`: Whether to create a light or dark theme. This affects the
            default values for some colors, such as the background.
        """
        # Primary palette
        if primary_color is None:
            primary_color = rio.Color.from_hex("01dffd")

        primary_palette = Palette(
            background=primary_color,
            background_variant=_derive_color(
                primary_color,
                0.08,
                bias_to_bright=-0.5,
            ),
            background_active=_derive_color(
                primary_color,
                0.15,
                bias_to_bright=0.6,
            ),
            foreground=(
                rio.Color.from_grey(0.1)
                if primary_color.perceived_brightness > 0.5
                else rio.Color.from_grey(0.9)
            ),
        )

        # Secondary palette
        if secondary_color is None:
            secondary_color = rio.Color.from_hex("0083ff")

        secondary_palette = Palette(
            background=secondary_color,
            background_variant=_derive_color(
                secondary_color,
                0.08,
                bias_to_bright=-0.3,
            ),
            background_active=_derive_color(
                secondary_color,
                0.15,
                bias_to_bright=0.6,
            ),
            foreground=(
                rio.Color.from_grey(0.1)
                if secondary_color.perceived_brightness > 0.75
                else rio.Color.from_grey(0.9)
            ),
        )

        # Background palette
        if background_color is None:
            if mode == "light":
                background_color = rio.Color.from_grey(1.00).blend(
                    primary_color, 0.05
                )
            else:
                background_color = rio.Color.from_grey(0.08).blend(
                    primary_color, 0.02
                )

        if text_color is None:
            neutral_and_background_text_color = (
                rio.Color.from_grey(0.3)
                if background_color.perceived_brightness > 0.5
                else rio.Color.from_grey(0.7)
            )
        else:
            neutral_and_background_text_color = text_color

        del text_color

        background_palette = Palette(
            background=background_color,
            background_variant=_derive_color(
                background_color,
                0.15,
                bias_to_bright=-0.15,
                target_color=primary_color,
            ),
            background_active=_derive_color(
                background_color,
                0.3,
                bias_to_bright=0.15,
                target_color=primary_color,
            ),
            foreground=neutral_and_background_text_color,
        )

        # Neutral palette
        #
        # This one is similar to the background palette, but with a slightly
        # different shade.
        if neutral_color is None:
            neutral_color = background_palette.background_variant

        neutral_palette = Palette(
            background=neutral_color,
            background_variant=_derive_color(
                neutral_color,
                0.3,
                bias_to_bright=-0.15,
                target_color=primary_color,
            ),
            background_active=_derive_color(
                neutral_color,
                0.5,
                bias_to_bright=0.15,
                target_color=primary_color,
            ),
            foreground=neutral_and_background_text_color,
        )

        # HUD palette
        if hud_color is None:
            hud_color = rio.Color.from_grey(0.1)

        hud_palette = Palette(
            background=hud_color,
            background_variant=_derive_color(
                hud_color,
                0.08,
            ),
            background_active=_derive_color(
                hud_color,
                0.15,
            ),
            foreground=(
                rio.Color.from_grey(0.1)
                if hud_color.perceived_brightness > 0.5
                else rio.Color.from_grey(0.9)
            ),
        )

        # Keep the disabled palette subdued. It's not meant to be perfectly
        # readable
        if disabled_color is None:
            disabled_color = rio.Color.from_grey(0.6)

        disabled_palette = Palette(
            background=disabled_color,
            background_variant=_derive_color(disabled_color, 0.20),
            background_active=_derive_color(disabled_color, 0.30),
            foreground=_derive_color(disabled_color, 0.4),
        )

        # Shadow color
        if mode == "light":
            shadow_color = rio.Color.from_rgb(0.1, 0.1, 0.4, 0.3)
        else:
            shadow_color = rio.Color.from_rgb(0, 0, 0, 0.4)

        # Semantic colors
        if success_color is None:
            success_color = rio.Color.from_hex("1e8e3e")

        if warning_color is None:
            warning_color = rio.Color.from_hex("f9a825")

        if danger_color is None:
            danger_color = rio.Color.from_hex("b3261e")

        success_palette = _make_semantic_palette(success_color)
        warning_palette = _make_semantic_palette(warning_color)
        danger_palette = _make_semantic_palette(danger_color)

        # Colorful headings can be a problem when the primary color is similar
        # to the background/neutral color. If the `color_headings` argument is
        # set to `auto`, disable coloring if the colors are close.
        if heading_fill == "auto":
            brightness1 = primary_palette.background.perceived_brightness
            brightness2 = background_palette.background.perceived_brightness

            heading_fill = (
                "primary" if abs(brightness1 - brightness2) > 0.3 else "plain"
            )

        if heading_fill == "primary":
            heading_fill = primary_color
        elif heading_fill == "plain":
            heading_fill = neutral_and_background_text_color
        else:
            heading_fill = heading_fill

        # Text styles
        heading1_style = rio.TextStyle(
            font=font,
            fill=heading_fill,
            font_size=2.3,
        )
        heading2_style = heading1_style.replace(font_size=1.7)
        heading3_style = heading1_style.replace(font_size=1.2)
        text_style = heading1_style.replace(
            font_size=1,
            fill=neutral_and_background_text_color,
        )

        # Instantiate the theme. `__init__` is blocked to prevent users from
        # doing something foolish. Work around that.
        return Theme._create_new(
            primary_palette=primary_palette,
            secondary_palette=secondary_palette,
            background_palette=background_palette,
            neutral_palette=neutral_palette,
            hud_palette=hud_palette,
            disabled_palette=disabled_palette,
            success_palette=success_palette,
            warning_palette=warning_palette,
            danger_palette=danger_palette,
            corner_radius_small=corner_radius_small,
            corner_radius_medium=corner_radius_medium,
            corner_radius_large=corner_radius_large,
            shadow_color=shadow_color,
            monospace_font=monospace_font,
            heading1_style=heading1_style,
            heading2_style=heading2_style,
            heading3_style=heading3_style,
            text_style=text_style,
        )

    @classmethod
    def pair_from_colors(
        cls,
        *,
        primary_color: rio.Color | None = None,
        secondary_color: rio.Color | None = None,
        background_color: rio.Color | None = None,
        neutral_color: rio.Color | None = None,
        hud_color: rio.Color | None = None,
        disabled_color: rio.Color | None = None,
        success_color: rio.Color | None = None,
        warning_color: rio.Color | None = None,
        danger_color: rio.Color | None = None,
        corner_radius_small: float = 0.4,
        corner_radius_medium: float = 0.8,
        corner_radius_large: float = 1.8,
        text_color: rio.Color
        | tuple[rio.Color | None, rio.Color | None]
        | None = None,
        font: text_style_module.Font = text_style_module.Font.ROBOTO,
        monospace_font: text_style_module.Font = text_style_module.Font.ROBOTO_MONO,
        heading_fill: t.Literal["primary", "plain", "auto"]
        | text_style_module._TextFill = "auto",
    ) -> tuple[Theme, Theme]:
        """
        This function is very similar to `from_colors`, but it returns two
        themes: A light and a dark one. When applying two themes to your app,
        Rio will automatically switch between them based on the user's system
        preferences.

        ```python
        # Create a theme pair
        themes = rio.Theme.pair_from_colors(
            # Configure your theme here
        )

        # And apply them to your app
        app = rio.App(
            theme=themes,
            # ...
        )
        ```

        ## Parameters

        `primary_color`: The main color of your app. This color will be used to
            tint the background and by some large components to fill large
            spaces with color

        `secondary_color`: A color that nicely complements the primary color. It
            is often used by small components such as buttons and switches.

        `background_color`: The app's background color. This should be a neutral
            color that doesn't distract from the content.

        `neutral_color`: Similar to the background color, it is also used for
            neutral areas. It should however be slightly different, allowing you
            to create a visual hierarchy. This is the default color of large
            elements such as cards.

        `hud_color`: Used for elements that pop over the content, such as
            tooltips.

        `disabled_color`: Used by insensitive components to indicate that they
            are not interactive. Typically a shade of grey.

        `success_color`: A color to give positive feedback the user. Typically
            a shade of green.

        `warning_color`: A color to indicate that something might be wrong, but
            isn't critical. Typically orange.

        `danger_color`: A color to indicate that something is wrong and needs
            immediate attention. Typically a shade of red.

        `corner_radius_small`: The corner radius of small components such as
            text inputs

        `corner_radius_medium`: The corner radius of medium-sized components,
            such as small cards.

        `corner_radius_large`: The corner radius of large components, such as
            large cards and dialogs.

        `heading_fill`: The fill to use for headings. This allows you to specify
            a more interesting color, or even a gradient. If set to `"auto"`,
            Rio will automatically switch between the primary color and a plain
            text color based on legibility.

            This only affects headings in background and neutral contexts.

        `text_color`: The default text color to use for regular text. Please
            note that this only applies to text in a neutral or background
            context. Text that's e.g. placed on a `rio.Card` with
            `color="primary"` will use a different color to ensure legibility.

            You can also specify a tuple of two colors to use different text
            colors for light and dark themes.

        `font`: The default font to use when no other is specified.

        `monospace_font`: The font to use for monospace text, such as code.
        """
        if not isinstance(text_color, tuple):
            text_color = (text_color, text_color)

        themes = list[Theme]()
        for mode, text_color in zip(("light", "dark"), text_color):
            themes.append(
                cls.from_colors(
                    primary_color=primary_color,
                    secondary_color=secondary_color,
                    background_color=background_color,
                    neutral_color=neutral_color,
                    hud_color=hud_color,
                    disabled_color=disabled_color,
                    success_color=success_color,
                    warning_color=warning_color,
                    danger_color=danger_color,
                    corner_radius_small=corner_radius_small,
                    corner_radius_medium=corner_radius_medium,
                    corner_radius_large=corner_radius_large,
                    font=font,
                    monospace_font=monospace_font,
                    heading_fill=heading_fill,
                    text_color=text_color,
                    mode=mode,
                )
            )

        return tuple(themes)  # type: ignore (length mismatch)

    def text_color_for(self, color: rio.Color) -> rio.Color:
        """
        Given the color of a background, return a legible text color to use on
        top of it.
        """
        return _derive_color(color, offset=0.8)

    def _serialize_colorset(self, color: color.ColorSet) -> Jsonable:
        # A colorset more-or-less translates to a switcheroo. Thus, this
        # function needs to provide all information needed to create one.

        # If the color is a string, just pass it through. In this case there is
        # an actual switcheroo which can be used.
        if isinstance(color, str):
            return color

        # TODO: If the color is identical/very similar to a theme color use the
        #       corresponding switcheroo.

        # Otherwise create all the necessary variables to emulate a switcheroo.
        palette = Palette.from_color(color)

        return {
            "localBg": palette.background.rgba,
            "localBgVariant": palette.background_variant.rgba,
            "localBgActive": palette.background_active.rgba,
            "localFg": palette.foreground.rgba,
        }

    @property
    def is_light_theme(self) -> bool:
        return self.background_palette.background.perceived_brightness >= 0.5

    @property
    def primary_color(self) -> rio.Color:
        return self.primary_palette.background

    @property
    def secondary_color(self) -> rio.Color:
        return self.secondary_palette.background

    @property
    def background_color(self) -> rio.Color:
        return self.background_palette.background

    @property
    def neutral_color(self) -> rio.Color:
        return self.neutral_palette.background

    @property
    def hud_color(self) -> rio.Color:
        return self.hud_palette.background

    @property
    def disabled_color(self) -> rio.Color:
        return self.disabled_palette.background

    @property
    def success_color(self) -> rio.Color:
        return self.success_palette.background

    @property
    def warning_color(self) -> rio.Color:
        return self.warning_palette.background

    @property
    def danger_color(self) -> rio.Color:
        return self.danger_palette.background

    @property
    def font(self) -> text_style_module.Font:
        assert (
            self.text_style.font is not None
        ), f"The theme's text style must have a font set"
        return self.text_style.font



from __future__ import annotations

import typing as t

import rio

from .. import components as comps


# <code>
class TicTacToePage(rio.Component):
    # The contents of all fields. Each field can contain an X, an O, or be
    # empty. The initial state is an empty board.
    fields: list[t.Literal["X", "O", ""]] = [""] * 9

    def build(self) -> rio.Component:
        # Spawn components for the fields
        field_components: list[rio.Component] = []

        for index, field in enumerate(self.fields):
            field_components.append(
                comps.Field(
                    value=field,
                )
            )

        # Arrange all components in a grid
        return rio.Grid(
            field_components[0:3],
            field_components[3:6],
            field_components[6:9],
            row_spacing=1,
            column_spacing=1,
            align_x=0.5,
        )


# </code>



from __future__ import annotations

import asyncio
import typing as t

import playwright.async_api
import playwright.sync_api
import uvicorn

import rio.app_server
import rio.data_models
from rio.app_server import FastapiServer
from rio.components.text import Text
from rio.debug.layouter import Layouter
from rio.session import Session
from rio.utils import choose_free_port

__all__ = ["verify_layout", "cleanup"]


layouter_factory: LayouterFactory | None = None


async def verify_layout(build: t.Callable[[], rio.Component]) -> Layouter:
    """
    Rio contains two layout implementations: One on the client side, which
    determines the real layout of components, and a second one on the server
    side which is used entirely for testing.

    This function verifies that the results from the two layouters are the same.
    """
    global layouter_factory

    if layouter_factory is None:
        layouter_factory = LayouterFactory()
        await layouter_factory.start()

    layouter = await layouter_factory.create_layouter(build)

    for component_id, layout_should in layouter._layouts_should.items():
        layout_is = layouter._layouts_are[component_id]

        differences = list[str]()
        for attribute in rio.data_models.ComponentLayout.__annotations__:
            # Not all attributes are meant to be compared
            if attribute == "parent_id":
                continue

            value_should = getattr(layout_should, attribute)
            value_is = getattr(layout_is, attribute)

            difference = abs(value_is - value_should)
            if difference > 0.2:
                differences.append(f"{attribute}: {value_is} != {value_should}")

        if differences:
            component = layouter.get_component_by_id(component_id)
            raise ValueError(
                f"Layout of component {component} is incorrect:\n- "
                + "\n- ".join(differences)
            )

    return layouter


async def cleanup() -> None:
    if layouter_factory is not None:
        await layouter_factory.stop()


class LayouterFactory:
    """
    This class is designed to efficiently create many `Layouter` objects for
    different GUIs. Starting a web server and a browser every time you need a
    `Layouter` has very high overhead, so this class re-uses the existing ones
    when possible.
    """

    def __init__(self) -> None:
        self._port = choose_free_port("localhost")

        self._app = rio.App(
            build=rio.Spacer,
            # JS reports incorrect sizes and positions for hidden elements, and
            # so the tests end up failing because of the icon in the connection
            # lost popup. I think it's because icons have a fixed size, but JS
            # reports the size as 0x0. So we'll get rid of the icon.
            build_connection_lost_message=build_connection_lost_message,
        )
        self._app_server: FastapiServer | None = None
        self._uvicorn_server: uvicorn.Server | None = None
        self._uvicorn_serve_task: asyncio.Task[None] | None = None
        self._playwright_context = None
        self._browser = None

    async def start(self) -> None:
        await self._start_browser()
        await self._start_uvicorn_server()

    async def stop(self) -> None:
        if self._browser is not None:
            await self._browser.close()

        if self._playwright_context is not None:
            await self._playwright_context.__aexit__()

        if self._uvicorn_server is not None:
            self._uvicorn_server.should_exit = True

            assert self._uvicorn_serve_task is not None
            await self._uvicorn_serve_task

    async def create_layouter(
        self, build: t.Callable[[], rio.Component]
    ) -> Layouter:
        self._app._build = build
        session, page = await self._create_session()

        # FIXME: Give the client some time to process the layout
        await asyncio.sleep(0.5)

        layouter = await Layouter.create(session)

        await page.close()
        await session._close(close_remote_session=False)

        return layouter

    async def _start_uvicorn_server(self) -> None:
        server_is_ready_event = asyncio.Event()
        loop = asyncio.get_running_loop()

        def set_server_ready_event() -> None:
            loop.call_soon_threadsafe(server_is_ready_event.set)

        self._app_server = FastapiServer(
            self._app,
            debug_mode=False,
            running_in_window=False,
            internal_on_app_start=set_server_ready_event,
            base_url=None,
        )

        config = uvicorn.Config(
            self._app_server,
            port=self._port,
            log_level="critical",
        )
        self._uvicorn_server = uvicorn.Server(config)

        current_task: asyncio.Task = asyncio.current_task()  # type: ignore
        self._uvicorn_serve_task = asyncio.create_task(
            self._run_uvicorn(current_task)
        )

        await server_is_ready_event.wait()

        # Just because uvicorn says it's ready doesn't mean it's actually ready.
        # Give it a bit more time.
        await asyncio.sleep(1)

    async def _run_uvicorn(self, test_task: asyncio.Task) -> None:
        assert self._uvicorn_server is not None

        try:
            await self._uvicorn_server.serve()
        except BaseException as error:
            test_task.cancel(f"Uvicorn server crashed: {error}")

    async def _start_browser(self) -> None:
        self._playwright_context = playwright.async_api.async_playwright()

        playwright_obj = await self._playwright_context.__aenter__()

        try:
            browser = await playwright_obj.chromium.launch()
        except Exception:
            raise Exception(
                "Playwright cannot launch chromium. Please execute the"
                " following command:\n"
                "playwright install --with-deps chromium\n"
                "(If you're using a virtual environment, activate it first.)"
            ) from None

        # With default settings, playwright gets detected as a crawler. So we
        # need to emulate a real device.
        self._browser = await browser.new_context(
            **playwright_obj.devices["Desktop Chrome"]
        )

    async def _create_session(self) -> tuple[Session, t.Any]:
        assert (
            self._app_server is not None
        ), "Uvicorn isn't running for some reason"
        assert self._browser is not None

        assert (
            not self._app_server.sessions
        ), f"App server still has sessions?! {self._app_server.sessions}"

        page = await self._browser.new_page()
        await page.goto(f"http://localhost:{self._port}")

        while not self._app_server.sessions:
            await asyncio.sleep(0.1)

        return self._app_server.sessions[0], page


def build_connection_lost_message() -> Text:
    return rio.Text("Connection Lost")



from __future__ import annotations

# <additional-imports>
import datetime
from dataclasses import field

import rio

from ..data_models import TodoItem

# </additional-imports>


# <component>
class NewTodoItemInput(rio.Component):
    """
    A component that lets the user create a new `TodoItem`.

    Includes a `TextInput` for the title and a button. Pressing the button or
    the "Enter" key will create a new `TodoItem`.
    """

    on_input: rio.EventHandler[TodoItem]

    # This is a private field that we need for an attribute binding. We don't
    # want to create a constructor parameter for this field, so we set
    # `init=False`.
    _title: str = field(init=False, default="")

    async def _on_confirm(
        self, _: rio.TextInputConfirmEvent | None = None
    ) -> None:
        # No title entered? Then we won't do anything
        if not self._title:
            return

        # Create the new `TodoItem` and call our `on_input` event handler
        new_todo_item = TodoItem(
            title=self._title,
            creation_time=datetime.datetime.now().astimezone(),
        )
        await self.call_event_handler(self.on_input, new_todo_item)

        # Reset the `TextInput`
        self._title = ""

    def build(self) -> rio.Component:
        return rio.Row(
            rio.TextInput(
                label="Enter a new todo item",
                text=self.bind()._title,
                on_confirm=self._on_confirm,
                grow_x=True,
            ),
            rio.IconButton(
                "material/add",
                min_size=2.5,
                on_press=self._on_confirm,
            ),
            spacing=1,
        )


# </component>



import rio


# <counter>
# <class-header>
class Counter(rio.Component):
    value: int = 0
    # </class-header>

    # <event-handlers>
    def _on_increment(self) -> None:
        self.value += 1

    def _on_decrement(self) -> None:
        self.value -= 1

    def _on_reset(self) -> None:
        self.value = 0

    # </event-handlers>

    # <build>
    def build(self):
        return rio.Column(
            rio.Row(
                rio.IconButton(
                    "material/remove",
                    on_press=self._on_decrement,
                    min_size=3,
                ),
                rio.Text(
                    str(self.value),
                    justify="center",
                    style="heading1",
                    min_width=5,
                ),
                rio.IconButton(
                    "material/add",
                    on_press=self._on_increment,
                    min_size=3,
                ),
                spacing=3,
            ),
            rio.Button(
                "Reset",
                style="minor",
                on_press=self._on_reset,
            ),
            spacing=2,
            align_x=0.5,
            align_y=0.5,
        )


# </build>
# </counter>


# <app>
app = rio.App(
    pages=[
        rio.ComponentPage("Home", "", Counter),
    ],
)
# </app>

# <run>
app.run_in_browser()
# </run>



import rio

from ...project_config import RioProjectConfig

__all__ = ["ProjectPage"]


class ProjectPage(rio.Component):
    project: RioProjectConfig | None = None

    def __post_init__(self) -> None:
        try:
            self.project = RioProjectConfig.try_locate_and_load()
        except FileNotFoundError:
            pass

    def build(self) -> rio.Component:
        # No project
        if self.project is None:
            return rio.Column(
                rio.Icon(
                    "material/error", min_width=4, min_height=4, fill="danger"
                ),
                rio.Text(
                    "Couldn't find your project files. Do you have a `rio.toml` file?",
                    overflow="wrap",
                ),
            )
        else:
            return ProjectComponent(self.project)


class ProjectComponent(rio.Component):
    project: RioProjectConfig

    async def _on_change_app_type(self, event: rio.DropdownChangeEvent) -> None:
        self.project.app_type = event.value
        self.project.write()
        await self.force_refresh()

    def build(self) -> rio.Component:
        project_name = self.project.project_directory.name.strip().capitalize()

        return rio.Column(
            rio.Text(
                project_name,
                style="heading2",
                justify="left",
            ),
            rio.Text(
                str(self.project.project_directory),
                style="dim",
                margin_bottom=2,
                justify="left",
            ),
            rio.Text(
                "Rio can create both apps and websites. Apps will launch in a separate window, while websites will launch in your browser. Which type is your project?",
                overflow="wrap",
            ),
            rio.Dropdown(
                label="Type",
                options={
                    "App": "app",
                    "Website": "website",
                },
            ),
            margin=1,
            align_y=0,
            spacing=0.5,
        )
